#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import errno
import os
import shutil
import sys
from datetime import timedelta
from io import BytesIO
from tempfile import mkdtemp
from unittest import mock

import boto3
import pytest
from moto import mock_aws

from airflow import DAG
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.operators.s3 import (
    S3CopyObjectOperator,
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
    S3DeleteBucketTaggingOperator,
    S3DeleteObjectsOperator,
    S3FileTransformOperator,
    S3GetBucketTaggingOperator,
    S3ListOperator,
    S3ListPrefixesOperator,
    S3PutBucketTaggingOperator,
)
from airflow.providers.common.compat.openlineage.facet import (
    Dataset,
    LifecycleStateChange,
    LifecycleStateChangeDatasetFacet,
    PreviousIdentifier,
)
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.openlineage.extractors import OperatorLineage
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.taskinstance import create_task_instance, render_template_fields
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_1_PLUS
from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if AIRFLOW_V_3_1_PLUS:
    from airflow.sdk.timezone import datetime, utcnow
else:
    from airflow.utils.timezone import datetime, utcnow  # type: ignore[attr-defined,no-redef]

BUCKET_NAME = os.environ.get("BUCKET_NAME", "test-airflow-bucket")
S3_KEY = "test-airflow-key"
TAG_SET = [{"Key": "Color", "Value": "Green"}]


class TestS3CreateBucketOperator:
    def setup_method(self):
        self.create_bucket_operator = S3CreateBucketOperator(
            task_id="test-s3-create-bucket-operator",
            bucket_name=BUCKET_NAME,
        )

    @mock_aws
    @mock.patch.object(S3Hook, "create_bucket")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_bucket_exist(self, mock_check_for_bucket, mock_create_bucket):
        mock_check_for_bucket.return_value = True
        # execute s3 bucket create operator
        self.create_bucket_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        mock_create_bucket.assert_not_called()

    @mock_aws
    @mock.patch.object(S3Hook, "create_bucket")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_not_bucket_exist(self, mock_check_for_bucket, mock_create_bucket):
        mock_check_for_bucket.return_value = False
        # execute s3 bucket create operator
        self.create_bucket_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        mock_create_bucket.assert_called_once_with(bucket_name=BUCKET_NAME, region_name=None)

    def test_template_fields(self):
        validate_template_fields(self.create_bucket_operator)


class TestS3DeleteBucketOperator:
    def setup_method(self):
        self.delete_bucket_operator = S3DeleteBucketOperator(
            task_id="test-s3-delete-operator",
            bucket_name=BUCKET_NAME,
        )

    @mock_aws
    @mock.patch.object(S3Hook, "delete_bucket")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_bucket_exist(self, mock_check_for_bucket, mock_delete_bucket):
        mock_check_for_bucket.return_value = True
        # execute s3 bucket delete operator
        self.delete_bucket_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        mock_delete_bucket.assert_called_once_with(bucket_name=BUCKET_NAME, force_delete=False)

    @mock_aws
    @mock.patch.object(S3Hook, "delete_bucket")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_not_bucket_exist(self, mock_check_for_bucket, mock_delete_bucket):
        mock_check_for_bucket.return_value = False
        # execute s3 bucket delete operator
        self.delete_bucket_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        mock_delete_bucket.assert_not_called()

    def test_template_fields(self):
        validate_template_fields(self.delete_bucket_operator)


class TestS3GetBucketTaggingOperator:
    def setup_method(self):
        self.get_bucket_tagging_operator = S3GetBucketTaggingOperator(
            task_id="test-s3-get-bucket-tagging-operator",
            bucket_name=BUCKET_NAME,
        )

    @mock_aws
    @mock.patch.object(S3Hook, "get_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_bucket_exist(self, mock_check_for_bucket, get_bucket_tagging):
        mock_check_for_bucket.return_value = True
        # execute s3 get bucket tagging operator
        self.get_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        get_bucket_tagging.assert_called_once_with(BUCKET_NAME)

    @mock_aws
    @mock.patch.object(S3Hook, "get_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_not_bucket_exist(self, mock_check_for_bucket, get_bucket_tagging):
        mock_check_for_bucket.return_value = False
        # execute s3 get bucket tagging operator
        self.get_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        get_bucket_tagging.assert_not_called()

    def test_template_fields(self):
        validate_template_fields(self.get_bucket_tagging_operator)


class TestS3PutBucketTaggingOperator:
    def setup_method(self):
        self.put_bucket_tagging_operator = S3PutBucketTaggingOperator(
            task_id="test-s3-put-bucket-tagging-operator",
            tag_set=TAG_SET,
            bucket_name=BUCKET_NAME,
        )

    @mock_aws
    @mock.patch.object(S3Hook, "put_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_bucket_exist(self, mock_check_for_bucket, put_bucket_tagging):
        mock_check_for_bucket.return_value = True
        # execute s3 put bucket tagging operator
        self.put_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        put_bucket_tagging.assert_called_once_with(
            key=None, value=None, tag_set=TAG_SET, bucket_name=BUCKET_NAME
        )

    @mock_aws
    @mock.patch.object(S3Hook, "put_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_not_bucket_exist(self, mock_check_for_bucket, put_bucket_tagging):
        mock_check_for_bucket.return_value = False
        # execute s3 put bucket tagging operator
        self.put_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        put_bucket_tagging.assert_not_called()

    def test_template_fields(self):
        validate_template_fields(self.put_bucket_tagging_operator)


class TestS3DeleteBucketTaggingOperator:
    def setup_method(self):
        self.delete_bucket_tagging_operator = S3DeleteBucketTaggingOperator(
            task_id="test-s3-delete-bucket-tagging-operator",
            bucket_name=BUCKET_NAME,
        )

    @mock_aws
    @mock.patch.object(S3Hook, "delete_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_bucket_exist(self, mock_check_for_bucket, delete_bucket_tagging):
        mock_check_for_bucket.return_value = True
        # execute s3 delete bucket tagging operator
        self.delete_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        delete_bucket_tagging.assert_called_once_with(BUCKET_NAME)

    @mock_aws
    @mock.patch.object(S3Hook, "delete_bucket_tagging")
    @mock.patch.object(S3Hook, "check_for_bucket")
    def test_execute_if_not_bucket_exist(self, mock_check_for_bucket, delete_bucket_tagging):
        mock_check_for_bucket.return_value = False
        # execute s3 delete bucket tagging operator
        self.delete_bucket_tagging_operator.execute({})
        mock_check_for_bucket.assert_called_once_with(BUCKET_NAME)
        delete_bucket_tagging.assert_not_called()

    def test_template_fields(self):
        validate_template_fields(self.delete_bucket_tagging_operator)


class TestS3FileTransformOperator:
    def setup_method(self):
        self.content = b"input"
        self.bucket = "bucket"
        self.input_key = "foo"
        self.output_key = "bar"
        self.bio = BytesIO(self.content)
        self.tmp_dir = mkdtemp(prefix="test_tmpS3FileTransform_")
        self.transform_script = os.path.join(self.tmp_dir, "transform.py")
        os.mknod(self.transform_script)

    def teardown_method(self):
        try:
            shutil.rmtree(self.tmp_dir)
        except OSError as e:
            # ENOENT - no such file or directory
            if e.errno != errno.ENOENT:
                raise e

    @mock.patch("subprocess.Popen")
    @mock.patch.object(S3FileTransformOperator, "log")
    @mock_aws
    def test_execute_with_transform_script(self, mock_log, mock_popen):
        process_output = [b"Foo", b"Bar", b"Baz"]
        self.mock_process(mock_popen, process_output=process_output)
        input_path, output_path = self.s3_paths()

        op = S3FileTransformOperator(
            source_s3_key=input_path,
            dest_s3_key=output_path,
            transform_script=self.transform_script,
            replace=True,
            task_id="task_id",
        )
        op.execute(None)

        mock_log.info.assert_has_calls(
            [mock.call(line.decode(sys.getdefaultencoding())) for line in process_output]
        )

    @mock.patch("subprocess.Popen")
    @mock_aws
    def test_execute_with_failing_transform_script(self, mock_popen):
        self.mock_process(mock_popen, return_code=42)
        input_path, output_path = self.s3_paths()

        op = S3FileTransformOperator(
            source_s3_key=input_path,
            dest_s3_key=output_path,
            transform_script=self.transform_script,
            replace=True,
            task_id="task_id",
        )

        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        assert str(ctx.value) == "Transform script failed: 42"

    @mock.patch("subprocess.Popen")
    @mock_aws
    def test_execute_with_transform_script_args(self, mock_popen):
        self.mock_process(mock_popen, process_output=[b"Foo", b"Bar", b"Baz"])
        input_path, output_path = self.s3_paths()
        script_args = ["arg1", "arg2"]

        op = S3FileTransformOperator(
            source_s3_key=input_path,
            dest_s3_key=output_path,
            transform_script=self.transform_script,
            script_args=script_args,
            replace=True,
            task_id="task_id",
        )
        op.execute(None)

        assert script_args == mock_popen.call_args.args[0][3:]

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.select_key", return_value="input")
    @mock_aws
    def test_execute_with_select_expression(self, mock_select_key):
        input_path, output_path = self.s3_paths()
        select_expression = "SELECT * FROM s3object s"
        input_serialization = None
        output_serialization = None

        op = S3FileTransformOperator(
            source_s3_key=input_path,
            dest_s3_key=output_path,
            select_expression=select_expression,
            replace=True,
            task_id="task_id",
        )
        op.execute(None)

        mock_select_key.assert_called_once_with(
            key=input_path,
            expression=select_expression,
            input_serialization=input_serialization,
            output_serialization=output_serialization,
        )

        conn = boto3.client("s3")
        result = conn.get_object(Bucket=self.bucket, Key=self.output_key)
        assert self.content == result["Body"].read()

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.select_key", return_value="input")
    @mock_aws
    def test_execute_with_select_expression_and_serialization_config(self, mock_select_key):
        input_path, output_path = self.s3_paths()
        select_expression = "SELECT * FROM s3object s"
        select_expr_serialization_config = {
            "input_serialization": {"CSV": {}},
            "output_serialization": {"CSV": {}},
        }

        op = S3FileTransformOperator(
            source_s3_key=input_path,
            dest_s3_key=output_path,
            select_expression=select_expression,
            select_expr_serialization_config=select_expr_serialization_config,
            replace=True,
            task_id="task_id",
        )
        op.execute(None)

        input_serialization = select_expr_serialization_config.get("input_serialization")
        output_serialization = select_expr_serialization_config.get("output_serialization")
        mock_select_key.assert_called_once_with(
            key=input_path,
            expression=select_expression,
            input_serialization=input_serialization,
            output_serialization=output_serialization,
        )

        conn = boto3.client("s3")
        result = conn.get_object(Bucket=self.bucket, Key=self.output_key)
        assert self.content == result["Body"].read()

    def test_get_openlineage_facets_on_start(self):
        expected_input = Dataset(
            namespace=f"s3://{self.bucket}",
            name=self.input_key,
        )
        expected_output = Dataset(
            namespace=f"s3://{self.bucket}",
            name=self.output_key,
        )

        op = S3FileTransformOperator(
            task_id="test",
            source_s3_key=f"s3://{self.bucket}/{self.input_key}",
            dest_s3_key=f"s3://{self.bucket}/{self.output_key}",
        )

        lineage = op.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == expected_input
        assert lineage.outputs[0] == expected_output

    @staticmethod
    def mock_process(mock_popen, return_code=0, process_output=None):
        mock_proc = mock.MagicMock()
        mock_proc.returncode = return_code
        mock_proc.stdout.readline.side_effect = process_output or []
        mock_proc.wait.return_value = None
        mock_popen.return_value.__enter__.return_value = mock_proc

    def s3_paths(self):
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.bucket)
        conn.upload_fileobj(Bucket=self.bucket, Key=self.input_key, Fileobj=self.bio)

        s3_url = "s3://{0}/{1}"
        input_path = s3_url.format(self.bucket, self.input_key)
        output_path = s3_url.format(self.bucket, self.output_key)

        return input_path, output_path

    def test_template_fields(self):
        operator = S3FileTransformOperator(
            source_s3_key="test/key",
            dest_s3_key="test/key",
            transform_script=self.transform_script,
            replace=True,
            task_id="task_id",
        )
        validate_template_fields(operator)


class TestS3ListOperator:
    def test_execute(self):
        operator = S3ListOperator(
            task_id="test-s3-list-operator",
            bucket=BUCKET_NAME,
            prefix="TEST",
            delimiter=".csv",
        )
        operator.hook = mock.MagicMock()
        operator.hook.list_keys.return_value = ["TEST1.csv", "TEST2.csv", "TEST3.csv"]

        files = operator.execute(None)

        operator.hook.list_keys.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            prefix="TEST",
            delimiter=".csv",
            apply_wildcard=False,
        )
        assert sorted(files) == sorted(["TEST1.csv", "TEST2.csv", "TEST3.csv"])

    def test_template_fields(self):
        operator = S3ListOperator(
            task_id="test-s3-list-operator",
            bucket=BUCKET_NAME,
            prefix="TEST",
            delimiter=".csv",
        )
        validate_template_fields(operator)


class TestS3ListPrefixesOperator:
    def test_execute(self):
        operator = S3ListPrefixesOperator(
            task_id="test-s3-list-prefixes-operator", bucket=BUCKET_NAME, prefix="test/", delimiter="/"
        )
        operator.hook = mock.MagicMock()
        operator.hook.list_prefixes.return_value = ["test/"]

        subfolders = operator.execute(None)

        operator.hook.list_prefixes.assert_called_once_with(
            bucket_name=BUCKET_NAME, prefix="test/", delimiter="/"
        )
        assert subfolders == ["test/"]

    def test_template_fields(self):
        operator = S3ListPrefixesOperator(
            task_id="test-s3-list-prefixes-operator", bucket=BUCKET_NAME, prefix="test/", delimiter="/"
        )
        validate_template_fields(operator)


class TestS3CopyObjectOperator:
    def setup_method(self):
        self.source_bucket = "bucket1"
        self.source_key = "path1/data.txt"
        self.dest_bucket = "bucket2"
        self.dest_key = "path2/data_copy.txt"

    @mock_aws
    def test_s3_copy_object_arg_combination_1(self):
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.source_bucket)
        conn.create_bucket(Bucket=self.dest_bucket)
        conn.upload_fileobj(Bucket=self.source_bucket, Key=self.source_key, Fileobj=BytesIO(b"input"))

        # there should be nothing found before S3CopyObjectOperator is executed
        assert "Contents" not in conn.list_objects(Bucket=self.dest_bucket, Prefix=self.dest_key)

        op = S3CopyObjectOperator(
            task_id="test_task_s3_copy_object",
            source_bucket_key=self.source_key,
            source_bucket_name=self.source_bucket,
            dest_bucket_key=self.dest_key,
            dest_bucket_name=self.dest_bucket,
        )
        op.execute(None)

        objects_in_dest_bucket = conn.list_objects(Bucket=self.dest_bucket, Prefix=self.dest_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.dest_key

    @mock_aws
    def test_s3_copy_object_arg_combination_2(self):
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.source_bucket)
        conn.create_bucket(Bucket=self.dest_bucket)
        conn.upload_fileobj(Bucket=self.source_bucket, Key=self.source_key, Fileobj=BytesIO(b"input"))

        # there should be nothing found before S3CopyObjectOperator is executed
        assert "Contents" not in conn.list_objects(Bucket=self.dest_bucket, Prefix=self.dest_key)

        source_key_s3_url = f"s3://{self.source_bucket}/{self.source_key}"
        dest_key_s3_url = f"s3://{self.dest_bucket}/{self.dest_key}"
        op = S3CopyObjectOperator(
            task_id="test_task_s3_copy_object",
            source_bucket_key=source_key_s3_url,
            dest_bucket_key=dest_key_s3_url,
        )
        op.execute(None)

        objects_in_dest_bucket = conn.list_objects(Bucket=self.dest_bucket, Prefix=self.dest_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.dest_key

    def test_get_openlineage_facets_on_start_combination_1(self):
        expected_input = Dataset(
            namespace=f"s3://{self.source_bucket}",
            name=self.source_key,
        )
        expected_output = Dataset(
            namespace=f"s3://{self.dest_bucket}",
            name=self.dest_key,
        )

        op = S3CopyObjectOperator(
            task_id="test",
            source_bucket_name=self.source_bucket,
            source_bucket_key=self.source_key,
            dest_bucket_name=self.dest_bucket,
            dest_bucket_key=self.dest_key,
        )

        lineage = op.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == expected_input
        assert lineage.outputs[0] == expected_output

    def test_get_openlineage_facets_on_start_combination_2(self):
        expected_input = Dataset(
            namespace=f"s3://{self.source_bucket}",
            name=self.source_key,
        )
        expected_output = Dataset(
            namespace=f"s3://{self.dest_bucket}",
            name=self.dest_key,
        )

        source_key_s3_url = f"s3://{self.source_bucket}/{self.source_key}"
        dest_key_s3_url = f"s3://{self.dest_bucket}/{self.dest_key}"

        op = S3CopyObjectOperator(
            task_id="test",
            source_bucket_key=source_key_s3_url,
            dest_bucket_key=dest_key_s3_url,
        )

        lineage = op.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == expected_input
        assert lineage.outputs[0] == expected_output

    def test_template_fields(self):
        operator = S3CopyObjectOperator(
            task_id="test_task_s3_copy_object",
            source_bucket_key=self.source_key,
            source_bucket_name=self.source_bucket,
            dest_bucket_key=self.dest_key,
            dest_bucket_name=self.dest_bucket,
        )
        validate_template_fields(operator)


@mock_aws
class TestS3DeleteObjectsOperator:
    def test_s3_delete_single_object(self):
        bucket = "testbucket"
        key = "path/data.txt"

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        conn.upload_fileobj(Bucket=bucket, Key=key, Fileobj=BytesIO(b"input"))

        # The object should be detected before the DELETE action is taken
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key)
        assert len(objects_in_dest_bucket["Contents"]) == 1
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key

        op = S3DeleteObjectsOperator(task_id="test_task_s3_delete_single_object", bucket=bucket, keys=key)
        op.execute(None)

        # There should be no object found in the bucket created earlier
        assert "Contents" not in conn.list_objects(Bucket=bucket, Prefix=key)

    def test_s3_delete_multiple_objects(self):
        bucket = "testbucket"
        key_pattern = "path/data"
        n_keys = 3
        keys = [key_pattern + str(i) for i in range(n_keys)]

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        for k in keys:
            conn.upload_fileobj(Bucket=bucket, Key=k, Fileobj=BytesIO(b"input"))

        # The objects should be detected before the DELETE action is taken
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key_pattern)
        assert len(objects_in_dest_bucket["Contents"]) == n_keys
        assert sorted(x["Key"] for x in objects_in_dest_bucket["Contents"]) == sorted(keys)

        op = S3DeleteObjectsOperator(task_id="test_task_s3_delete_multiple_objects", bucket=bucket, keys=keys)
        op.execute(None)

        # There should be no object found in the bucket created earlier
        assert "Contents" not in conn.list_objects(Bucket=bucket, Prefix=key_pattern)

    @pytest.mark.db_test
    def test_dates_from_template(self, session, testing_dag_bundle):
        """Specifically test for dates passed from templating that could be strings"""
        bucket = "testbucket"
        key_pattern = "path/data"
        n_keys = 3
        keys = [key_pattern + str(i) for i in range(n_keys)]

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        for k in keys:
            conn.upload_fileobj(Bucket=bucket, Key=k, Fileobj=BytesIO(b"input"))

        logical_date = utcnow()
        dag = DAG("test_dag", start_date=datetime(2020, 1, 1), schedule=timedelta(days=1))
        # use macros.ds_add since it returns a string, not a date
        op = S3DeleteObjectsOperator(
            task_id="XXXXXXXXXXXXXXXXXXXXXXX",
            bucket=bucket,
            from_datetime="{{ macros.ds_add(ds, -1) }}",
            to_datetime="{{ macros.ds_add(ds, 1) }}",
            dag=dag,
        )
        if hasattr(DagRun, "execution_date"):  # Airflow 2.x.
            dag_run = DagRun(
                dag_id=dag.dag_id,
                execution_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                logical_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
                run_after=utcnow(),
            )
        if AIRFLOW_V_3_0_PLUS:
            from airflow.models.dag_version import DagVersion

            sync_dag_to_db(dag)
            dag_version = DagVersion.get_latest_version(dag.dag_id)
            ti = create_task_instance(task=op, run_id="test", dag_version_id=dag_version.id)
        else:
            ti = TaskInstance(task=op)
        ti.dag_run = dag_run
        rendered = render_template_fields(ti, op)
        rendered.execute(None)
        assert "Contents" not in conn.list_objects(Bucket=bucket)

    def test_s3_delete_from_to_datetime(self):
        bucket = "testbucket"
        key_pattern = "path/data"
        n_keys = 3
        keys = [key_pattern + str(i) for i in range(n_keys)]

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        for k in keys:
            conn.upload_fileobj(Bucket=bucket, Key=k, Fileobj=BytesIO(b"input"))

        # The objects should be detected before the DELETE action is taken
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket)
        assert len(objects_in_dest_bucket["Contents"]) == n_keys
        assert sorted(x["Key"] for x in objects_in_dest_bucket["Contents"]) == sorted(keys)

        now = utcnow()
        from_datetime = now.replace(year=now.year - 1)
        to_datetime = now.replace(year=now.year + 1)

        op = S3DeleteObjectsOperator(
            task_id="test_task_s3_delete_prefix",
            bucket=bucket,
            from_datetime=from_datetime,
            to_datetime=to_datetime,
        )
        op.execute(None)

        # There should be no object found in the bucket created earlier
        assert "Contents" not in conn.list_objects(Bucket=bucket)

    def test_s3_delete_prefix(self):
        bucket = "testbucket"
        key_pattern = "path/data"
        n_keys = 3
        keys = [key_pattern + str(i) for i in range(n_keys)]

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        for k in keys:
            conn.upload_fileobj(Bucket=bucket, Key=k, Fileobj=BytesIO(b"input"))

        # The objects should be detected before the DELETE action is taken
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key_pattern)
        assert len(objects_in_dest_bucket["Contents"]) == n_keys
        assert sorted(x["Key"] for x in objects_in_dest_bucket["Contents"]) == sorted(keys)

        op = S3DeleteObjectsOperator(task_id="test_task_s3_delete_prefix", bucket=bucket, prefix=key_pattern)
        op.execute(None)

        # There should be no object found in the bucket created earlier
        assert "Contents" not in conn.list_objects(Bucket=bucket, Prefix=key_pattern)

    def test_s3_delete_empty_list(self):
        bucket = "testbucket"
        key_of_test = "path/data.txt"
        keys = []

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        conn.upload_fileobj(Bucket=bucket, Key=key_of_test, Fileobj=BytesIO(b"input"))

        # The object should be detected before the DELETE action is tested
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key_of_test)
        assert len(objects_in_dest_bucket["Contents"]) == 1
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

        op = S3DeleteObjectsOperator(task_id="test_s3_delete_empty_list", bucket=bucket, keys=keys)
        op.execute(None)

        # The object found in the bucket created earlier should still be there
        assert len(objects_in_dest_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

    def test_s3_delete_empty_string(self):
        bucket = "testbucket"
        key_of_test = "path/data.txt"
        keys = ""

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        conn.upload_fileobj(Bucket=bucket, Key=key_of_test, Fileobj=BytesIO(b"input"))

        # The object should be detected before the DELETE action is tested
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key_of_test)
        assert len(objects_in_dest_bucket["Contents"]) == 1
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

        op = S3DeleteObjectsOperator(task_id="test_s3_delete_empty_string", bucket=bucket, keys=keys)
        op.execute(None)

        # The object found in the bucket created earlier should still be there
        assert len(objects_in_dest_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

    @pytest.mark.parametrize(
        ("keys", "prefix", "from_datetime", "to_datetime"),
        [
            pytest.param("path/data.txt", "path/data", None, None, id="single-key-and-prefix"),
            pytest.param(["path/data.txt"], "path/data", None, None, id="multiple-keys-and-prefix"),
            pytest.param(
                ["path/data.txt"],
                "path/data",
                datetime(1992, 3, 8, 18, 52, 51),
                None,
                id="keys-prefix-and-from_datetime",
            ),
            pytest.param(
                ["path/data.txt"],
                "path/data",
                datetime(1992, 3, 8, 18, 52, 51),
                datetime(1993, 3, 8, 18, 52, 51),
                id="keys-prefix-and-from-to_datetime",
            ),
            pytest.param(None, None, None, None, id="all-none"),
        ],
    )
    def test_validate_keys_and_filters_in_constructor(self, keys, prefix, from_datetime, to_datetime):
        with pytest.raises(
            AirflowException,
            match=r"Either keys or at least one of prefix, from_datetime, to_datetime should be set.",
        ):
            S3DeleteObjectsOperator(
                task_id="test_validate_keys_and_prefix_in_constructor",
                bucket="foo-bar-bucket",
                keys=keys,
                prefix=prefix,
                from_datetime=from_datetime,
                to_datetime=to_datetime,
            )

    @pytest.mark.parametrize(
        ("keys", "prefix", "from_datetime", "to_datetime"),
        [
            pytest.param("path/data.txt", "path/data", None, None, id="single-key-and-prefix"),
            pytest.param(["path/data.txt"], "path/data", None, None, id="multiple-keys-and-prefix"),
            pytest.param(
                ["path/data.txt"],
                "path/data",
                datetime(1992, 3, 8, 18, 52, 51),
                None,
                id="keys-prefix-and-from_datetime",
            ),
            pytest.param(
                ["path/data.txt"],
                "path/data",
                datetime(1992, 3, 8, 18, 52, 51),
                datetime(1993, 3, 8, 18, 52, 51),
                id="keys-prefix-and-from-to_datetime",
            ),
            pytest.param(None, None, None, None, id="all-none"),
        ],
    )
    def test_validate_keys_and_prefix_in_execute(self, keys, prefix, from_datetime, to_datetime):
        bucket = "testbucket"
        key_of_test = "path/data.txt"

        conn = boto3.client("s3")
        conn.create_bucket(Bucket=bucket)
        conn.upload_fileobj(Bucket=bucket, Key=key_of_test, Fileobj=BytesIO(b"input"))

        # Set valid values for constructor, and change them later for emulate rendering template
        op = S3DeleteObjectsOperator(
            task_id="test_validate_keys_and_prefix_in_execute",
            bucket=bucket,
            keys="keys-exists",
            prefix=None,
        )
        op.keys = keys
        op.prefix = prefix
        op.from_datetime = from_datetime
        op.to_datetime = to_datetime

        # The object should be detected before the DELETE action is tested
        objects_in_dest_bucket = conn.list_objects(Bucket=bucket, Prefix=key_of_test)
        assert len(objects_in_dest_bucket["Contents"]) == 1
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

        with pytest.raises(
            AirflowException,
            match=r"Either keys or at least one of prefix, from_datetime, to_datetime should be set.",
        ):
            op.execute(None)

        # The object found in the bucket created earlier should still be there
        assert len(objects_in_dest_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == key_of_test

    @pytest.mark.parametrize("keys", ("path/data.txt", ["path/data.txt"]))
    def test_get_openlineage_facets_on_complete_single_object(self, keys):
        bucket = "testbucket"
        expected_input = Dataset(
            namespace=f"s3://{bucket}",
            name="path/data.txt",
            facets={
                "lifecycleStateChange": LifecycleStateChangeDatasetFacet(
                    lifecycleStateChange=LifecycleStateChange.DROP.value,
                    previousIdentifier=PreviousIdentifier(
                        namespace=f"s3://{bucket}",
                        name="path/data.txt",
                    ),
                )
            },
        )

        op = S3DeleteObjectsOperator(task_id="test_task_s3_delete_single_object", bucket=bucket, keys=keys)
        op.hook = mock.MagicMock()
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0] == expected_input

    def test_get_openlineage_facets_on_complete_multiple_objects(self):
        bucket = "testbucket"
        keys = ["path/data1.txt", "path/data2.txt"]
        expected_inputs = [
            Dataset(
                namespace=f"s3://{bucket}",
                name="path/data1.txt",
                facets={
                    "lifecycleStateChange": LifecycleStateChangeDatasetFacet(
                        lifecycleStateChange=LifecycleStateChange.DROP.value,
                        previousIdentifier=PreviousIdentifier(
                            namespace=f"s3://{bucket}",
                            name="path/data1.txt",
                        ),
                    )
                },
            ),
            Dataset(
                namespace=f"s3://{bucket}",
                name="path/data2.txt",
                facets={
                    "lifecycleStateChange": LifecycleStateChangeDatasetFacet(
                        lifecycleStateChange=LifecycleStateChange.DROP.value,
                        previousIdentifier=PreviousIdentifier(
                            namespace=f"s3://{bucket}",
                            name="path/data2.txt",
                        ),
                    )
                },
            ),
        ]

        op = S3DeleteObjectsOperator(task_id="test_task_s3_delete_single_object", bucket=bucket, keys=keys)
        op.hook = mock.MagicMock()
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 2
        assert lineage.inputs == expected_inputs

    @pytest.mark.parametrize("keys", ("", []))
    @mock.patch("airflow.providers.amazon.aws.operators.s3.S3Hook")
    def test_get_openlineage_facets_on_complete_no_objects(self, mock_hook, keys):
        op = S3DeleteObjectsOperator(
            task_id="test_task_s3_delete_single_object", bucket="testbucket", keys=keys
        )
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)
        assert lineage == OperatorLineage()

    def test_template_fields(self):
        operator = S3DeleteObjectsOperator(
            task_id="test_task_s3_delete_single_object", bucket="test-bucket", keys="test/file.csv"
        )
        validate_template_fields(operator)


class TestS3CreateObjectOperator:
    @mock.patch.object(S3Hook, "load_string")
    def test_execute_if_data_is_string(self, mock_load_string):
        data = "data"
        operator = S3CreateObjectOperator(
            task_id="test-s3-operator",
            s3_bucket=BUCKET_NAME,
            s3_key=S3_KEY,
            data=data,
        )
        operator.execute(None)

        mock_load_string.assert_called_once_with(data, S3_KEY, BUCKET_NAME, False, False, None, None, None)

    @mock.patch.object(S3Hook, "load_bytes")
    def test_execute_if_data_is_bytes(self, mock_load_bytes):
        data = b"data"
        operator = S3CreateObjectOperator(
            task_id="test-s3-create-object-operator",
            s3_bucket=BUCKET_NAME,
            s3_key=S3_KEY,
            data=data,
        )
        operator.execute(None)

        mock_load_bytes.assert_called_once_with(data, S3_KEY, BUCKET_NAME, False, False, None)

    @mock.patch.object(S3Hook, "load_string")
    def test_execute_if_s3_bucket_not_provided(self, mock_load_string):
        data = "data"
        operator = S3CreateObjectOperator(
            task_id="test-s3-create-object-operator",
            s3_key=f"s3://{BUCKET_NAME}/{S3_KEY}",
            data=data,
        )
        operator.execute(None)

        mock_load_string.assert_called_once_with(data, S3_KEY, BUCKET_NAME, False, False, None, None, None)

    @pytest.mark.parametrize(("bucket", "key"), (("bucket", "file.txt"), (None, "s3://bucket/file.txt")))
    def test_get_openlineage_facets_on_start(self, bucket, key):
        expected_output = Dataset(
            namespace="s3://bucket",
            name="file.txt",
        )

        op = S3CreateObjectOperator(task_id="test", s3_bucket=bucket, s3_key=key, data="test")

        lineage = op.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0] == expected_output

    def test_template_fields(self):
        operator = S3CreateObjectOperator(task_id="test", s3_bucket="bucket", s3_key="key", data="test")
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import random
import string
import uuid

import pytest

from airflow.providers.amazon.aws.utils.identifiers import generate_uuid
from airflow.providers.amazon.version_compat import NOTSET


class TestGenerateUuid:
    @pytest.fixture(
        autouse=True,
        params=[
            pytest.param(NOTSET, id="default-namespace"),
            pytest.param(uuid.UUID(int=42), id="custom-namespace"),
        ],
    )
    def setup_namespace(self, request):
        self.default_namespace = request.param is NOTSET
        self.namespace = uuid.NAMESPACE_OID if self.default_namespace else request.param
        self.kwargs = {"namespace": self.namespace} if not self.default_namespace else {}

    def test_deterministic(self):
        """Test that result is deterministic and a valid UUID object"""
        args = ["".join(random.choices(string.ascii_letters, k=random.randint(3, 13))) for _ in range(100)]
        result = generate_uuid(*args, **self.kwargs)
        assert result == generate_uuid(*args, **self.kwargs)
        assert uuid.UUID(result).version == 5, "Should generate UUID v5"

    def test_nil_uuid(self):
        """Test that result of single None are NIL UUID, regardless namespace."""
        assert generate_uuid(None, **self.kwargs) == "00000000-0000-0000-0000-000000000000"

    def test_single_uuid_value(self):
        """Test that result of single not None value are the same as uuid5."""
        assert generate_uuid("", **self.kwargs) == str(uuid.uuid5(self.namespace, ""))
        assert generate_uuid("Airflow", **self.kwargs) == str(uuid.uuid5(self.namespace, "Airflow"))

    def test_multiple_none_value(self):
        """Test that result of single None are NIL UUID, regardless of namespace."""
        multi_none = generate_uuid(None, None, **self.kwargs)
        assert multi_none != "00000000-0000-0000-0000-000000000000"
        assert uuid.UUID(multi_none).version == 5

        # Test that None values not skipped
        assert generate_uuid(None, "1", None, **self.kwargs) != generate_uuid("1", **self.kwargs)
        assert generate_uuid(None, "1", **self.kwargs) != generate_uuid("1", **self.kwargs)
        assert generate_uuid("1", None, **self.kwargs) != generate_uuid("1", **self.kwargs)

    def test_no_args_value(self):
        with pytest.raises(ValueError, match="Expected at least 1 argument"):
            generate_uuid(**self.kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime as dt
import json
from unittest import mock

import pytest
from botocore.exceptions import ClientError
from semver import VersionInfo

from airflow.executors.base_executor import BaseExecutor
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.amazon.aws.executors.aws_lambda import lambda_executor
from airflow.providers.amazon.aws.executors.aws_lambda.lambda_executor import AwsLambdaExecutor
from airflow.providers.amazon.aws.executors.aws_lambda.utils import CONFIG_GROUP_NAME, AllLambdaConfigKeys
from airflow.providers.common.compat.sdk import AirflowException
from airflow.utils.state import TaskInstanceState
from airflow.version import version as airflow_version_str

from tests_common.test_utils.compat import timezone
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

airflow_version = VersionInfo(*map(int, airflow_version_str.split(".")[:3]))

DEFAULT_QUEUE_URL = "queue-url"
DEFAULT_DLQ_URL = "dlq-url"
DEFAULT_FUNCTION_NAME = "function-name"


@pytest.fixture
def set_env_vars():
    overrides: dict[tuple[str, str], str] = {
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.AWS_CONN_ID): "aws_default",
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.REGION_NAME): "us-west-1",
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.FUNCTION_NAME): DEFAULT_FUNCTION_NAME,
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.QUEUE_URL): DEFAULT_QUEUE_URL,
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.DLQ_URL): DEFAULT_DLQ_URL,
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.QUALIFIER): "1",
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.MAX_INVOKE_ATTEMPTS): "3",
        (CONFIG_GROUP_NAME, AllLambdaConfigKeys.CHECK_HEALTH_ON_STARTUP): "True",
    }
    with conf_vars(overrides):
        yield


@pytest.fixture
def mock_airflow_key():
    def _key():
        key_mock = mock.Mock()
        # Use a "random" value (memory id of the mock obj) so each key serializes uniquely
        key_mock._asdict = mock.Mock(return_value={"mock_key": id(key_mock)})
        return key_mock

    return _key


def _generate_mock_cmd():
    return ["airflow", "tasks", "run", "dag_id", "task_id", "run_id", "--local"]


# The following two fixtures look different because no existing test
# cares if they have unique values, so the same value is always used.
@pytest.fixture
def mock_cmd():
    return _generate_mock_cmd()


@pytest.fixture
def mock_executor(set_env_vars) -> AwsLambdaExecutor:
    """Mock Lambda to a repeatable starting state.."""
    executor = AwsLambdaExecutor()
    executor.IS_BOTO_CONNECTION_HEALTHY = True

    # Replace boto3 clients with mocks
    lambda_mock = mock.Mock(spec=executor.lambda_client)
    lambda_mock.invoke.return_value = {"StatusCode": 0, "failures": []}
    executor.lambda_client = lambda_mock

    sqs_mock = mock.Mock(spec=executor.sqs_client)
    sqs_mock.receive_message.return_value = {"Messages": []}
    executor.sqs_client = sqs_mock

    return executor


class TestAwsLambdaExecutor:
    @mock.patch(
        "airflow.providers.amazon.aws.executors.aws_lambda.lambda_executor.AwsLambdaExecutor.change_state"
    )
    def test_execute(self, change_state_mock, mock_airflow_key, mock_executor, mock_cmd):
        """Test execution from end-to-end."""
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        assert len(mock_executor.pending_tasks) == 0
        mock_executor.execute_async(airflow_key, mock_cmd)
        assert len(mock_executor.pending_tasks) == 1

        mock_executor.attempt_task_runs()
        mock_executor.lambda_client.invoke.assert_called_once()
        payload = json.loads(mock_executor.lambda_client.invoke.call_args.kwargs["Payload"])
        assert payload["executor_config"] == {}

        # Task is stored in active worker.
        assert len(mock_executor.running_tasks) == 1
        assert json.dumps(airflow_key._asdict()) in mock_executor.running_tasks
        change_state_mock.assert_called_once_with(
            airflow_key, TaskInstanceState.RUNNING, ser_airflow_key, remove_running=False
        )

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3+")
    @mock.patch(
        "airflow.providers.amazon.aws.executors.aws_lambda.lambda_executor.AwsLambdaExecutor.change_state"
    )
    def test_task_sdk(self, change_state_mock, mock_airflow_key, mock_executor, mock_cmd):
        """Test task sdk execution from end-to-end."""
        from airflow.executors.workloads import ExecuteTask

        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())
        executor_config = {"config_key": "config_value"}

        workload = mock.Mock(spec=ExecuteTask)
        workload.ti = mock.Mock(spec=TaskInstance)
        workload.ti.key = airflow_key
        workload.ti.executor_config = executor_config
        ser_workload = json.dumps({"test_key": "test_value"})
        workload.model_dump_json.return_value = ser_workload

        mock_executor.queue_workload(workload, mock.Mock())

        assert mock_executor.queued_tasks[workload.ti.key] == workload
        assert len(mock_executor.pending_tasks) == 0
        assert len(mock_executor.running) == 0
        mock_executor._process_workloads([workload])
        assert len(mock_executor.queued_tasks) == 0
        assert len(mock_executor.running) == 1
        assert workload.ti.key in mock_executor.running
        assert len(mock_executor.pending_tasks) == 1
        assert mock_executor.pending_tasks[0].command == [
            "python",
            "-m",
            "airflow.sdk.execution_time.execute_workload",
            "--json-string",
            '{"test_key": "test_value"}',
        ]

        mock_executor.attempt_task_runs()
        mock_executor.lambda_client.invoke.assert_called_once()
        payload = json.loads(mock_executor.lambda_client.invoke.call_args.kwargs["Payload"])
        assert payload["executor_config"] == executor_config
        assert len(mock_executor.pending_tasks) == 0

        # Task is stored in active worker.
        assert len(mock_executor.running_tasks) == 1
        assert mock_executor.running_tasks[ser_airflow_key] == workload.ti.key
        change_state_mock.assert_called_once_with(
            workload.ti.key, TaskInstanceState.RUNNING, ser_airflow_key, remove_running=False
        )

    @mock.patch.object(lambda_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_success_execute_api_exception(self, mock_backoff, mock_executor, mock_cmd, mock_airflow_key):
        """Test what happens when Lambda throws an initial exception on invoke, but ultimately passes on retries."""
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        invoke_exception = Exception("Test exception")
        invoke_success = {"StatusCode": 0, "failures": []}
        mock_executor.lambda_client.invoke.side_effect = [invoke_exception, invoke_exception, invoke_success]
        mock_executor.execute_async(airflow_key, mock_cmd)
        expected_retry_count = 2

        # Fail 2 times
        for _ in range(expected_retry_count):
            mock_executor.attempt_task_runs()
            # Task is not stored in active workers.
            assert len(mock_executor.running_tasks) == 0

        # Pass in last attempt
        mock_executor.attempt_task_runs()
        assert len(mock_executor.pending_tasks) == 0
        assert ser_airflow_key in mock_executor.running_tasks
        assert mock_backoff.call_count == expected_retry_count
        for attempt_number in range(1, expected_retry_count):
            mock_backoff.assert_has_calls([mock.call(attempt_number)])

    def test_failed_execute_api_exception(self, mock_executor, mock_cmd, mock_airflow_key):
        """Test what happens when Lambda refuses to execute a task and throws an exception"""
        mock_airflow_key = mock_airflow_key()

        mock_executor.lambda_client.invoke.side_effect = Exception("Test exception")
        mock_executor.execute_async(mock_airflow_key, mock_cmd)

        # No matter what, don't schedule until invoke becomes successful.
        for _ in range(int(mock_executor.max_invoke_attempts) * 2):
            mock_executor.attempt_task_runs()
            # Task is not stored in running tasks
            assert len(mock_executor.running_tasks) == 0

    def test_failed_execute_creds_exception(self, mock_executor, mock_cmd, mock_airflow_key):
        """Test what happens when Lambda refuses to execute a task and throws an exception due to credentials"""
        airflow_key = mock_airflow_key()

        mock_executor.IS_BOTO_CONNECTION_HEALTHY = True
        mock_executor.execute_async(airflow_key, mock_cmd)
        assert mock_executor.pending_tasks[0].attempt_number == 1

        error_to_raise = ClientError(
            {"Error": {"Code": "ExpiredTokenException", "Message": "foobar"}}, "OperationName"
        )
        mock_executor.lambda_client.invoke.side_effect = error_to_raise

        # Sync will ultimately call attempt_task_runs, which is the code under test
        mock_executor.sync()

        # Task should end up back in the queue
        assert mock_executor.pending_tasks[0].key == airflow_key
        # The connection should get marked as unhealthy
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY
        # We retry on connections issues indefinitely, so the attempt number should be 1
        assert mock_executor.pending_tasks[0].attempt_number == 1

    def test_failed_execute_client_error_exception(self, mock_executor, mock_cmd, mock_airflow_key):
        """Test what happens when Lambda refuses to execute a task and throws an exception for non-credentials issue"""
        airflow_key = mock_airflow_key()
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = True
        mock_executor.execute_async(airflow_key, mock_cmd)
        assert mock_executor.pending_tasks[0].attempt_number == 1

        error_to_raise = ClientError(
            {"Error": {"Code": "RandomeError", "Message": "foobar"}}, "OperationName"
        )
        mock_executor.lambda_client.invoke.side_effect = error_to_raise

        # Sync will ultimately call attempt_task_runs, which is the code under test
        mock_executor.sync()

        # Task should end up back in the queue
        assert mock_executor.pending_tasks[0].key == airflow_key
        # The connection should stay marked as healthy because the error is something else
        assert mock_executor.IS_BOTO_CONNECTION_HEALTHY
        # Not a retry so increment attempts
        assert mock_executor.pending_tasks[0].attempt_number == 2

    @mock.patch.object(lambda_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_attempt_task_runs_attempts_when_tasks_fail(self, _, mock_executor):
        """
        Test case when all tasks fail to run.

        The executor should attempt each task exactly once per sync() iteration.
        It should preserve the order of tasks, and attempt each task up to
        `max_invoke_attempts` times before dropping the task.
        """
        airflow_keys = [
            TaskInstanceKey("a", "task_a", "c", 1, -1),
            TaskInstanceKey("a", "task_b", "c", 1, -1),
        ]
        airflow_cmd1 = _generate_mock_cmd()
        airflow_cmd2 = _generate_mock_cmd()
        commands = [airflow_cmd1, airflow_cmd2]

        failures = [Exception("Failure 1"), Exception("Failure 2")]

        mock_executor.execute_async(airflow_keys[0], commands[0])
        mock_executor.execute_async(airflow_keys[1], commands[1])

        assert len(mock_executor.pending_tasks) == 2
        assert len(mock_executor.running_tasks) == 0

        mock_executor.lambda_client.invoke.side_effect = failures
        mock_executor.attempt_task_runs()

        for i in range(2):
            payload = json.loads(mock_executor.lambda_client.invoke.call_args_list[i].kwargs["Payload"])
            assert airflow_keys[i].task_id in payload["task_key"]

        assert len(mock_executor.pending_tasks) == 2
        assert len(mock_executor.running_tasks) == 0

        mock_executor.lambda_client.invoke.call_args_list.clear()

        mock_executor.lambda_client.invoke.side_effect = failures
        mock_executor.attempt_task_runs()

        for i in range(2):
            payload = json.loads(mock_executor.lambda_client.invoke.call_args_list[i].kwargs["Payload"])
            assert airflow_keys[i].task_id in payload["task_key"]

        assert len(mock_executor.pending_tasks) == 2
        assert len(mock_executor.running_tasks) == 0

        mock_executor.lambda_client.invoke.call_args_list.clear()

        mock_executor.lambda_client.invoke.side_effect = failures
        mock_executor.attempt_task_runs()

        assert (
            len(mock_executor.pending_tasks) == 0
        )  # Pending now zero since we've had three failures to invoke
        assert len(mock_executor.running_tasks) == 0

        if airflow_version >= (2, 10, 0):
            events = [(x.event, x.task_id, x.try_number) for x in mock_executor._task_event_logs]
            assert events == [
                ("lambda invoke failure", "task_a", 1),
                ("lambda invoke failure", "task_b", 1),
            ]

    @mock.patch.object(lambda_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_attempt_task_runs_attempts_when_some_tasks_fal(self, _, mock_executor):
        """
        Test case when one task fail to run, others succeed, and a new task gets queued.

        """
        airflow_keys = [
            TaskInstanceKey("a", "task_a", "c", 1, -1),
            TaskInstanceKey("a", "task_b", "c", 1, -1),
        ]
        airflow_cmd1 = _generate_mock_cmd()
        airflow_cmd2 = _generate_mock_cmd()
        airflow_commands = [airflow_cmd1, airflow_cmd2]

        success_response = {"StatusCode": 0, "failures": []}

        responses = [Exception("Failure 1"), success_response]

        mock_executor.execute_async(airflow_keys[0], airflow_commands[0])
        mock_executor.execute_async(airflow_keys[1], airflow_commands[1])

        assert len(mock_executor.pending_tasks) == 2

        mock_executor.lambda_client.invoke.side_effect = responses
        mock_executor.attempt_task_runs()

        for i in range(2):
            payload = json.loads(mock_executor.lambda_client.invoke.call_args_list[i].kwargs["Payload"])
            assert airflow_keys[i].task_id in payload["task_key"]

        assert len(mock_executor.pending_tasks) == 1
        assert len(mock_executor.running_tasks) == 1

        mock_executor.lambda_client.invoke.call_args_list.clear()

        # queue new task
        airflow_keys[1] = TaskInstanceKey("a", "task_c", "c", 1, -1)
        airflow_commands[1] = _generate_mock_cmd()
        mock_executor.execute_async(airflow_keys[1], airflow_commands[1])

        assert len(mock_executor.pending_tasks) == 2
        # assert that the order of pending tasks is preserved i.e. the first task is 1st etc.
        assert mock_executor.pending_tasks[0].key == airflow_keys[0]
        assert mock_executor.pending_tasks[0].command == airflow_commands[0]

        responses = [Exception("Failure 1"), success_response]
        mock_executor.lambda_client.invoke.side_effect = responses
        mock_executor.attempt_task_runs()

        for i in range(2):
            payload = json.loads(mock_executor.lambda_client.invoke.call_args_list[i].kwargs["Payload"])
            assert airflow_keys[i].task_id in payload["task_key"]

        assert len(mock_executor.pending_tasks) == 1
        assert len(mock_executor.running_tasks) == 2

        mock_executor.lambda_client.invoke.call_args_list.clear()

        responses = [Exception("Failure 1")]
        mock_executor.lambda_client.invoke.side_effect = responses
        mock_executor.attempt_task_runs()

        payload = json.loads(mock_executor.lambda_client.invoke.call_args_list[0].kwargs["Payload"])
        assert airflow_keys[0].task_id in payload["task_key"]

        if airflow_version >= (2, 10, 0):
            events = [(x.event, x.task_id, x.try_number) for x in mock_executor._task_event_logs]
            assert events == [("lambda invoke failure", "task_a", 1)]

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_dlq(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        mock_executor.sqs_client.receive_message.side_effect = [
            {},  # First request from the results queue will be empty
            {
                # Second request from the DLQ will have a message
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": json.dumps(
                            {
                                "task_key": ser_airflow_key,
                                # DLQ messages will have the input (task_key, command) instead of return_code
                                "command": "command",
                            }
                        ),
                    }
                ]
            },
        ]

        mock_executor.sync_running_tasks()
        # Receive messages should be called twice
        assert mock_executor.sqs_client.receive_message.call_count == 2
        assert mock_executor.sqs_client.receive_message.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_QUEUE_URL,
            "MaxNumberOfMessages": 10,
        }

        assert mock_executor.sqs_client.receive_message.call_args_list[1].kwargs == {
            "QueueUrl": DEFAULT_DLQ_URL,
            "MaxNumberOfMessages": 10,
        }

        # Task is not stored in active workers.
        assert len(mock_executor.running_tasks) == 0
        success_mock.assert_not_called()
        fail_mock.assert_called_once()
        assert mock_executor.sqs_client.delete_message.call_count == 1
        assert mock_executor.sqs_client.delete_message.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_DLQ_URL,
            "ReceiptHandle": "receipt_handle",
        }

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_success(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        # Success message
        mock_executor.sqs_client.receive_message.return_value = {
            "Messages": [
                {
                    "ReceiptHandle": "receipt_handle",
                    "Body": json.dumps(
                        {
                            "task_key": ser_airflow_key,
                            "return_code": 0,
                        }
                    ),
                }
            ]
        }

        mock_executor.sync_running_tasks()
        mock_executor.sqs_client.receive_message.assert_called_once()
        assert mock_executor.sqs_client.receive_message.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_QUEUE_URL,
            "MaxNumberOfMessages": 10,
        }

        # Task is not stored in active workers.
        assert len(mock_executor.running_tasks) == 0
        # Task is immediately succeeded.
        success_mock.assert_called_once()
        fail_mock.assert_not_called()
        assert mock_executor.sqs_client.delete_message.call_count == 1
        assert mock_executor.sqs_client.delete_message.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_QUEUE_URL,
            "ReceiptHandle": "receipt_handle",
        }

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_fail(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        # Failure message
        mock_executor.sqs_client.receive_message.return_value = {
            "Messages": [
                {
                    "ReceiptHandle": "receipt_handle",
                    "Body": json.dumps(
                        {
                            "task_key": ser_airflow_key,
                            "return_code": 1,  # Non-zero return code, task failed
                        }
                    ),
                }
            ]
        }

        mock_executor.sync_running_tasks()
        mock_executor.sqs_client.receive_message.assert_called_once()

        # Task is not stored in active workers.
        assert len(mock_executor.running_tasks) == 0
        # Task is immediately succeeded.
        success_mock.assert_not_called()
        fail_mock.assert_called_once()
        assert mock_executor.sqs_client.delete_message.call_count == 1

    def test_sync_running_fail_bad_json(self, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        mock_executor.sqs_client.receive_message.side_effect = [
            {
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": "Banana",  # Body not json format
                    }
                ]
            },
            {},  # Second request from the DLQ will be empty
        ]

        mock_executor.sync_running_tasks()
        # Assert that the message is deleted if the message is not formatted as json
        assert mock_executor.sqs_client.receive_message.call_count == 2
        assert mock_executor.sqs_client.delete_message.call_count == 1

    def test_sync_running_fail_bad_format(self, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        mock_executor.sqs_client.receive_message.side_effect = [
            {
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": json.dumps(
                            {
                                "foo": "bar",  # Missing expected keys like "task_key"
                                "return_code": 1,  # Non-zero return code, task failed
                            }
                        ),
                    }
                ]
            },
            {},  # Second request from the DLQ will be empty
        ]

        mock_executor.sync_running_tasks()
        # Assert that the message is deleted if the message does not contain the expected keys
        assert mock_executor.sqs_client.receive_message.call_count == 2
        assert mock_executor.sqs_client.delete_message.call_count == 1

    def test_sync_running_fail_bad_format_dlq(self, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        # Failure message
        mock_executor.sqs_client.receive_message.side_effect = [
            {},  # First request from the results queue will be empty
            {
                # Second request from the DLQ will have a message
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": json.dumps(
                            {
                                "foo": "bar",  # Missing expected keys like "task_key"
                                "return_code": 1,
                            }
                        ),
                    }
                ]
            },
        ]

        mock_executor.sync_running_tasks()
        # Assert that the message is deleted if the message does not contain the expected keys
        assert mock_executor.sqs_client.receive_message.call_count == 2
        assert mock_executor.sqs_client.delete_message.call_count == 1

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_short_circuit(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        mock_executor.running_tasks.clear()
        # No running tasks, so we will short circuit

        mock_executor.sync_running_tasks()
        mock_executor.sqs_client.receive_message.assert_not_called()

        # Task is still stored in active workers.
        assert len(mock_executor.running_tasks) == 0
        # Task is immediately succeeded.
        success_mock.assert_not_called()
        fail_mock.assert_not_called()
        assert mock_executor.sqs_client.delete_message.call_count == 0

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_no_updates(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        # No messages, so we will not loop
        mock_executor.sqs_client.receive_message.return_value = {"Messages": []}

        mock_executor.sync_running_tasks()
        # Both the results queue and DLQ should have been checked
        assert mock_executor.sqs_client.receive_message.call_count == 2

        # Task is still stored in active workers.
        assert len(mock_executor.running_tasks) == 1
        # Task is immediately succeeded.
        success_mock.assert_not_called()
        fail_mock.assert_not_called()
        assert mock_executor.sqs_client.delete_message.call_count == 0

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_two_tasks_one_relevant(
        self, success_mock, fail_mock, mock_executor, mock_airflow_key
    ):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())
        airflow_key_2 = mock_airflow_key()
        ser_airflow_key_2 = json.dumps(airflow_key_2._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        mock_executor.running_tasks[ser_airflow_key_2] = airflow_key_2
        # Success message
        mock_executor.sqs_client.receive_message.side_effect = [
            {
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": json.dumps(
                            {
                                "task_key": ser_airflow_key,
                                "return_code": 0,
                            }
                        ),
                    }
                ]
            },
            {},  # No messages from DLQ
        ]

        mock_executor.sync_running_tasks()
        # Both the results queue and DLQ should have been checked
        assert mock_executor.sqs_client.receive_message.call_count == 2

        # One task left running
        assert len(mock_executor.running_tasks) == 1
        # Task one completed, task two is still running
        assert ser_airflow_key_2 in mock_executor.running_tasks
        # Task is immediately succeeded.
        success_mock.assert_called_once()
        fail_mock.assert_not_called()
        assert mock_executor.sqs_client.delete_message.call_count == 1

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync_running_unknown_task(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())
        airflow_key_2 = mock_airflow_key()
        ser_airflow_key_2 = json.dumps(airflow_key_2._asdict())

        mock_executor.running_tasks.clear()
        # Only add one of the tasks to the running list, the other will be unknown
        mock_executor.running_tasks[ser_airflow_key] = airflow_key

        # Receive the known task and unknown task
        known_task_receipt = "receipt_handle_known"
        unknown_task_receipt = "receipt_handle_unknown"
        mock_executor.sqs_client.receive_message.return_value = {
            "Messages": [
                {
                    "ReceiptHandle": known_task_receipt,
                    "Body": json.dumps(
                        {
                            "task_key": ser_airflow_key,
                            "return_code": 0,
                        }
                    ),
                },
                {
                    "ReceiptHandle": unknown_task_receipt,
                    "Body": json.dumps(
                        {
                            "task_key": ser_airflow_key_2,
                            "return_code": 0,
                        }
                    ),
                },
            ]
        }

        mock_executor.sync_running_tasks()
        mock_executor.sqs_client.receive_message.assert_called_once()

        # The known task is set to succeeded, unknown task is dropped
        assert len(mock_executor.running_tasks) == 0
        success_mock.assert_called_once()
        fail_mock.assert_not_called()
        # Only the known message from the queue should be deleted, the other should be marked as visible again
        assert mock_executor.sqs_client.delete_message.call_count == 1
        assert mock_executor.sqs_client.change_message_visibility.call_count == 1
        # The argument to delete_message should be the known task
        assert mock_executor.sqs_client.delete_message.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_QUEUE_URL,
            "ReceiptHandle": known_task_receipt,
        }
        # The change_message_visibility should be called with the unknown task
        assert mock_executor.sqs_client.change_message_visibility.call_args_list[0].kwargs == {
            "QueueUrl": DEFAULT_QUEUE_URL,
            "ReceiptHandle": unknown_task_receipt,
            "VisibilityTimeout": 0,
        }

    def test_start_no_check_health(self, mock_executor):
        mock_executor.check_health = mock.Mock()
        with conf_vars({(CONFIG_GROUP_NAME, AllLambdaConfigKeys.CHECK_HEALTH_ON_STARTUP): "False"}):
            mock_executor.start()

        assert mock_executor.check_health.call_count == 0

    def test_start_check_health_success(self, mock_executor):
        mock_executor.check_health = mock.Mock()
        with conf_vars({(CONFIG_GROUP_NAME, AllLambdaConfigKeys.CHECK_HEALTH_ON_STARTUP): "True"}):
            mock_executor.start()

        assert mock_executor.check_health.call_count == 1

    def test_start_check_health_fail(self, mock_executor):
        mock_executor.check_health = mock.Mock()
        mock_executor.check_health.side_effect = AirflowException("Test exception")
        with conf_vars({(CONFIG_GROUP_NAME, AllLambdaConfigKeys.CHECK_HEALTH_ON_STARTUP): "True"}):
            with pytest.raises(AirflowException):
                mock_executor.start()

        assert mock_executor.check_health.call_count == 1

    def test_check_health_success(self, mock_executor):
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sqs_client.get_queue_attributes.return_value = {
            "Attributes": {"ApproximateNumberOfMessages": 0}
        }
        mock_executor.lambda_client.get_function.return_value = {
            "Configuration": {
                "FunctionName": DEFAULT_FUNCTION_NAME,
                "State": "Active",
            }
        }
        mock_executor.check_health()
        assert mock_executor.sqs_client.get_queue_attributes.call_count == 2
        assert mock_executor.lambda_client.get_function.call_count == 1
        assert mock_executor.IS_BOTO_CONNECTION_HEALTHY

    def test_check_health_lambda_fails(self, mock_executor):
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sqs_client.get_queue_attributes.return_value = {
            "Attributes": {"ApproximateNumberOfMessages": 0}
        }
        mock_executor.lambda_client.get_function.return_value = ClientError(
            {"Error": {"Code": "ResourceNotFoundException", "Message": "foobar"}}, "OperationName"
        )

        with pytest.raises(AirflowException):
            mock_executor.check_health()
        assert mock_executor.lambda_client.get_function.call_count == 1
        # Lambda has already failed so SQS should not be called
        assert mock_executor.sqs_client.get_queue_attributes.call_count == 0
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY

    def test_check_health_sqs_fails(self, mock_executor):
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sqs_client.get_queue_attributes.return_value = ClientError(
            {"Error": {"Code": "ResourceNotFoundException", "Message": "foobar"}}, "OperationName"
        )
        mock_executor.lambda_client.get_function.return_value = {
            "Configuration": {
                "FunctionName": DEFAULT_FUNCTION_NAME,
                "State": "Active",
            }
        }
        with pytest.raises(AirflowException):
            mock_executor.check_health()
        assert mock_executor.lambda_client.get_function.call_count == 1
        # Lambda has already failed so SQS should not be called
        assert mock_executor.sqs_client.get_queue_attributes.call_count == 1
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY

    def test_check_health_sqs_results_queue_success_dlq_fails(self, mock_executor):
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sqs_client.get_queue_attributes.side_effect = [
            {"Attributes": {"ApproximateNumberOfMessages": 0}},
            ClientError(
                {"Error": {"Code": "ResourceNotFoundException", "Message": "foobar"}}, "OperationName"
            ),
        ]
        mock_executor.lambda_client.get_function.return_value = {
            "Configuration": {
                "FunctionName": DEFAULT_FUNCTION_NAME,
                "State": "Active",
            }
        }
        with pytest.raises(AirflowException):
            mock_executor.check_health()
        assert mock_executor.lambda_client.get_function.call_count == 1
        # Lambda has already failed so SQS should not be called
        assert mock_executor.sqs_client.get_queue_attributes.call_count == 2
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY

    def test_sync_already_unhealthy(self, mock_executor):
        # Something has set the connection to unhealthy (tested elsewhere)
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sync_running_tasks = mock.Mock()
        mock_executor.attempt_task_runs = mock.Mock()
        mock_executor.load_connections = mock.Mock()
        # Set the last connection reload to be more than 60 seconds ago so that we get a reload
        mock_executor.last_connection_reload = timezone.utcnow() - dt.timedelta(seconds=100)
        # We should not be able to sync
        mock_executor.sync()
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY
        mock_executor.sync_running_tasks.assert_not_called()
        mock_executor.attempt_task_runs.assert_not_called()
        mock_executor.load_connections.assert_called_once()

    def test_sync_already_unhealthy_then_repaired(self, mock_executor):
        # Something has set the connection to unhealthy (tested elsewhere)
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sync_running_tasks = mock.Mock()
        mock_executor.attempt_task_runs = mock.Mock()

        def check_health_side_effect():
            mock_executor.IS_BOTO_CONNECTION_HEALTHY = True

        mock_executor.check_health = mock.Mock(side_effect=check_health_side_effect)
        # Set the last connection reload to be more than 60 seconds ago so that we get a reload
        mock_executor.last_connection_reload = timezone.utcnow() - dt.timedelta(seconds=100)
        # Sync should repair itself and continue to call the sync methods
        mock_executor.sync()
        assert mock_executor.IS_BOTO_CONNECTION_HEALTHY
        mock_executor.sync_running_tasks.assert_called_once()
        mock_executor.attempt_task_runs.assert_called_once()

    @pytest.mark.parametrize(
        "error_code",
        [
            "ExpiredTokenException",
            "InvalidClientTokenId",
            "UnrecognizedClientException",
        ],
    )
    def test_sync_become_unhealthy_no_creds(self, error_code, mock_executor):
        # Something has set the connection to unhealthy (tested elsewhere)
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = True
        mock_executor.log.warning = mock.Mock()
        mock_executor.attempt_task_runs = mock.Mock()
        error_to_raise = ClientError({"Error": {"Code": error_code, "Message": "foobar"}}, "OperationName")
        mock_executor.sync_running_tasks = mock.Mock(side_effect=error_to_raise)

        # sync should catch the error and handle it, setting connection to unhealthy
        mock_executor.sync()
        assert not mock_executor.IS_BOTO_CONNECTION_HEALTHY
        mock_executor.sync_running_tasks.assert_called_once()
        mock_executor.attempt_task_runs.assert_not_called()
        # Check that the substring "AWS credentials are either missing or expired" was logged
        mock_executor.log.warning.assert_called_once()
        assert "AWS credentials are either missing or expired" in mock_executor.log.warning.call_args[0][0]

    def test_sync_exception(self, mock_executor):
        # Something has set the connection to unhealthy (tested elsewhere)
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = True
        mock_executor.log.exception = mock.Mock()
        mock_executor.attempt_task_runs = mock.Mock()
        mock_executor.sync_running_tasks = mock.Mock(side_effect=Exception())

        # sync should catch the error and log, don't kill scheduler by letting it raise up higher.
        mock_executor.sync()
        # Not a credentials error that we can tell, so connection stays healthy
        assert mock_executor.IS_BOTO_CONNECTION_HEALTHY
        mock_executor.sync_running_tasks.assert_called_once()
        mock_executor.attempt_task_runs.assert_not_called()
        # Check that the substring "AWS credentials are either missing or expired" was logged
        mock_executor.log.exception.assert_called_once()
        assert "An error occurred while syncing tasks" in mock_executor.log.exception.call_args[0][0]

    def test_try_adopt_task_instances(self, mock_executor, mock_airflow_key):
        """Test that executor can adopt orphaned task instances from a SchedulerJob shutdown event."""
        # airflow_key_1 = mock_airflow_key()
        airflow_key_1 = TaskInstanceKey("a", "task_a", "c", 1, -1)
        ser_airflow_key_1 = json.dumps(airflow_key_1._asdict())
        # airflow_key_2 = mock_airflow_key()
        airflow_key_2 = TaskInstanceKey("a", "task_b", "c", 1, -1)
        ser_airflow_key_2 = json.dumps(airflow_key_2._asdict())

        orphaned_tasks = [
            mock.Mock(spec=TaskInstance),
            mock.Mock(spec=TaskInstance),
            mock.Mock(spec=TaskInstance),
        ]
        orphaned_tasks[0].external_executor_id = ser_airflow_key_1
        orphaned_tasks[1].external_executor_id = ser_airflow_key_2
        orphaned_tasks[
            2
        ].external_executor_id = None  # One orphaned task has no external_executor_id, not adopted

        for task in orphaned_tasks:
            task.try_number = 1

        not_adopted_tasks = mock_executor.try_adopt_task_instances(orphaned_tasks)

        # Two of the three tasks should be adopted.
        assert len(orphaned_tasks) - 1 == len(mock_executor.running_tasks)
        assert ser_airflow_key_1 in mock_executor.running_tasks

        assert mock_executor.running_tasks[ser_airflow_key_1] == airflow_key_1
        assert ser_airflow_key_2 in mock_executor.running_tasks
        assert mock_executor.running_tasks[ser_airflow_key_2] == airflow_key_2

        # The remaining one task is unable to be adopted.
        assert len(not_adopted_tasks) == 1
        assert not_adopted_tasks[0] == orphaned_tasks[2]

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_end(self, success_mock, fail_mock, mock_executor, mock_airflow_key):
        """Test that executor can end successfully; waiting for all tasks to naturally exit."""
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        # First message is empty, so we loop again while waiting for tasks to finish
        mock_executor.sqs_client.receive_message.side_effect = [
            {},
            {},
            {
                "Messages": [
                    {
                        "ReceiptHandle": "receipt_handle",
                        "Body": json.dumps(
                            {
                                "task_key": ser_airflow_key,
                                "return_code": 0,
                            }
                        ),
                    }
                ]
            },
        ]
        mock_executor.end(heartbeat_interval=0)
        # Assert that the sqs_client mock method receive_message was called exactly twice
        assert mock_executor.sqs_client.receive_message.call_count == 3

        # Task is not stored in active workers.
        assert len(mock_executor.running_tasks) == 0
        success_mock.assert_called_once()
        fail_mock.assert_not_called()
        assert mock_executor.sqs_client.delete_message.call_count == 1

    @mock.patch("airflow.providers.amazon.aws.executors.aws_lambda.lambda_executor.timezone")
    def test_end_timeout(self, mock_timezone, mock_executor, mock_airflow_key):
        """Test that executor can end successfully; waiting for all tasks to naturally exit."""
        # Mock the sync method of the mock_executor object so we can count how many times it was called
        mock_executor.sync = mock.Mock()
        mock_executor.log.warning = mock.Mock()
        current_time = timezone.utcnow()
        mock_timezone.utcnow.side_effect = [
            current_time,
            current_time,
            current_time + dt.timedelta(seconds=5),
            current_time + dt.timedelta(seconds=10),
        ]

        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key

        with conf_vars({(CONFIG_GROUP_NAME, AllLambdaConfigKeys.END_WAIT_TIMEOUT): "5"}):
            mock_executor.end(heartbeat_interval=0)

        # Task is still stored in active workers.
        assert len(mock_executor.running_tasks) == 1
        assert mock_executor.sync.call_count == 2
        mock_executor.log.warning.assert_called_once_with(
            "Timed out waiting for tasks to finish. Some tasks may not be handled gracefully"
            " as the executor is force ending due to timeout."
        )

    def test_terminate(self, mock_executor, mock_airflow_key):
        """Test that executor can terminate successfully."""
        airflow_key = mock_airflow_key()
        ser_airflow_key = json.dumps(airflow_key._asdict())

        mock_executor.running_tasks.clear()
        mock_executor.running_tasks[ser_airflow_key] = airflow_key
        mock_executor.log.warning = mock.Mock()

        mock_executor.terminate()
        mock_executor.log.warning.assert_called_once_with(
            "Terminating Lambda executor. In-flight tasks cannot be stopped."
        )
        assert len(mock_executor.running_tasks) == 1



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import time

from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.common.compat.sdk import AirflowException


class QuickSightHook(AwsBaseHook):
    """
    Interact with Amazon QuickSight.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("quicksight") <QuickSight.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    NON_TERMINAL_STATES = {"INITIALIZED", "QUEUED", "RUNNING"}
    FAILED_STATES = {"FAILED"}

    def __init__(self, *args, **kwargs):
        super().__init__(client_type="quicksight", *args, **kwargs)

    def create_ingestion(
        self,
        data_set_id: str,
        ingestion_id: str,
        ingestion_type: str,
        wait_for_completion: bool = True,
        check_interval: int = 30,
        aws_account_id: str | None = None,
    ) -> dict:
        """
        Create and start a new SPICE ingestion for a dataset; refresh the SPICE datasets.

        .. seealso::
            - :external+boto3:py:meth:`QuickSight.Client.create_ingestion`

        :param data_set_id:  ID of the dataset used in the ingestion.
        :param ingestion_id: ID for the ingestion.
        :param ingestion_type: Type of ingestion: "INCREMENTAL_REFRESH"|"FULL_REFRESH"
        :param wait_for_completion: if the program should keep running until job finishes
        :param check_interval: the time interval in seconds which the operator
            will check the status of QuickSight Ingestion
        :param aws_account_id: An AWS Account ID, if set to ``None`` then use associated AWS Account ID.
        :return: Returns descriptive information about the created data ingestion
            having Ingestion ARN, HTTP status, ingestion ID and ingestion status.
        """
        aws_account_id = aws_account_id or self.account_id
        self.log.info("Creating QuickSight Ingestion for data set id %s.", data_set_id)
        try:
            create_ingestion_response = self.conn.create_ingestion(
                DataSetId=data_set_id,
                IngestionId=ingestion_id,
                IngestionType=ingestion_type,
                AwsAccountId=aws_account_id,
            )

            if wait_for_completion:
                self.wait_for_state(
                    aws_account_id=aws_account_id,
                    data_set_id=data_set_id,
                    ingestion_id=ingestion_id,
                    target_state={"COMPLETED"},
                    check_interval=check_interval,
                )
            return create_ingestion_response
        except Exception as general_error:
            self.log.error("Failed to run Amazon QuickSight create_ingestion API, error: %s", general_error)
            raise

    def get_status(self, aws_account_id: str | None, data_set_id: str, ingestion_id: str) -> str:
        """
        Get the current status of QuickSight Create Ingestion API.

        .. seealso::
            - :external+boto3:py:meth:`QuickSight.Client.describe_ingestion`

        :param aws_account_id: An AWS Account ID, if set to ``None`` then use associated AWS Account ID.
        :param data_set_id: QuickSight Data Set ID
        :param ingestion_id: QuickSight Ingestion ID
        :return: An QuickSight Ingestion Status
        """
        aws_account_id = aws_account_id or self.account_id
        try:
            describe_ingestion_response = self.conn.describe_ingestion(
                AwsAccountId=aws_account_id, DataSetId=data_set_id, IngestionId=ingestion_id
            )
            return describe_ingestion_response["Ingestion"]["IngestionStatus"]
        except KeyError as e:
            raise AirflowException(f"Could not get status of the Amazon QuickSight Ingestion: {e}")
        except ClientError as e:
            raise AirflowException(f"AWS request failed: {e}")

    def get_error_info(self, aws_account_id: str | None, data_set_id: str, ingestion_id: str) -> dict | None:
        """
        Get info about the error if any.

        :param aws_account_id: An AWS Account ID, if set to ``None`` then use associated AWS Account ID.
        :param data_set_id: QuickSight Data Set ID
        :param ingestion_id: QuickSight Ingestion ID
        :return: Error info dict containing the error type (key 'Type') and message (key 'Message')
            if available. Else, returns None.
        """
        aws_account_id = aws_account_id or self.account_id

        describe_ingestion_response = self.conn.describe_ingestion(
            AwsAccountId=aws_account_id, DataSetId=data_set_id, IngestionId=ingestion_id
        )
        # using .get() to get None if the key is not present, instead of an exception.
        return describe_ingestion_response["Ingestion"].get("ErrorInfo")

    def wait_for_state(
        self,
        aws_account_id: str | None,
        data_set_id: str,
        ingestion_id: str,
        target_state: set,
        check_interval: int,
    ):
        """
        Check status of a QuickSight Create Ingestion API.

        :param aws_account_id: An AWS Account ID, if set to ``None`` then use associated AWS Account ID.
        :param data_set_id: QuickSight Data Set ID
        :param ingestion_id: QuickSight Ingestion ID
        :param target_state: Describes the QuickSight Job's Target State
        :param check_interval: the time interval in seconds which the operator
            will check the status of QuickSight Ingestion
        :return: response of describe_ingestion call after Ingestion is done
        """
        aws_account_id = aws_account_id or self.account_id

        while True:
            status = self.get_status(aws_account_id, data_set_id, ingestion_id)
            self.log.info("Current status is %s", status)
            if status in self.FAILED_STATES:
                info = self.get_error_info(aws_account_id, data_set_id, ingestion_id)
                raise AirflowException(f"The Amazon QuickSight Ingestion failed. Error info: {info}")
            if status == "CANCELLED":
                raise AirflowException("The Amazon QuickSight SPICE ingestion cancelled!")
            if status not in self.NON_TERMINAL_STATES or status == target_state:
                break
            time.sleep(check_interval)

        self.log.info("QuickSight Ingestion completed")
        return status



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.models import DAG, DagRun, TaskInstance
from airflow.providers.amazon.aws.transfers.mongo_to_s3 import MongoToS3Operator
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.compat import timezone
from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.taskinstance import create_task_instance, render_template_fields
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

TASK_ID = "test_mongo_to_s3_operator"
MONGO_CONN_ID = "default_mongo"
AWS_CONN_ID = "default_s3"
MONGO_COLLECTION = "example_collection"
MONGO_QUERY = {"$lt": "{{ ts + 'Z' }}"}
S3_BUCKET = "example_bucket"
S3_KEY = "example_key"
COMPRESSION = None

DEFAULT_DATE = timezone.datetime(2017, 1, 1)
MOCK_MONGO_RETURN = [
    {"example_return_key_1": "example_return_value_1"},
    {"example_return_key_2": "example_return_value_2"},
]


class TestMongoToS3Operator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}

        self.dag = DAG("test_dag_id", schedule=None, default_args=args)

        self.mock_operator = MongoToS3Operator(
            task_id=TASK_ID,
            mongo_conn_id=MONGO_CONN_ID,
            aws_conn_id=AWS_CONN_ID,
            mongo_collection=MONGO_COLLECTION,
            mongo_query=MONGO_QUERY,
            s3_bucket=S3_BUCKET,
            s3_key=S3_KEY,
            dag=self.dag,
            compression=COMPRESSION,
        )

    def test_init(self):
        assert self.mock_operator.task_id == TASK_ID
        assert self.mock_operator.mongo_conn_id == MONGO_CONN_ID
        assert self.mock_operator.aws_conn_id == AWS_CONN_ID
        assert self.mock_operator.mongo_collection == MONGO_COLLECTION
        assert self.mock_operator.mongo_query == MONGO_QUERY
        assert self.mock_operator.s3_bucket == S3_BUCKET
        assert self.mock_operator.s3_key == S3_KEY
        assert self.mock_operator.compression == COMPRESSION

    def test_template_field_overrides(self):
        assert self.mock_operator.template_fields == (
            "s3_bucket",
            "s3_key",
            "mongo_query",
            "mongo_collection",
        )

    @pytest.mark.db_test
    def test_render_template(self, session, clean_dags_dagruns_and_dagbundles, testing_dag_bundle):
        if AIRFLOW_V_3_0_PLUS:
            from airflow.models.dag_version import DagVersion

            sync_dag_to_db(self.dag)
            dag_version = DagVersion.get_latest_version(self.mock_operator.dag_id)
            dag_run = DagRun(
                dag_id=self.mock_operator.dag_id,
                logical_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
                run_after=timezone.utcnow(),
            )
            ti = create_task_instance(self.mock_operator, run_id="test", dag_version_id=dag_version.id)
        else:
            dag_run = DagRun(
                dag_id=self.mock_operator.dag_id,
                execution_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
            ti = TaskInstance(task=self.mock_operator)
        ti.dag_run = dag_run
        render_template_fields(ti, self.mock_operator)
        expected_rendered_template = {"$lt": "2017-01-01T00:00:00+00:00Z"}
        assert expected_rendered_template == getattr(self.mock_operator, "mongo_query")

    @mock.patch("airflow.providers.amazon.aws.transfers.mongo_to_s3.MongoHook")
    @mock.patch("airflow.providers.amazon.aws.transfers.mongo_to_s3.S3Hook")
    def test_execute(self, mock_s3_hook, mock_mongo_hook):
        operator = self.mock_operator

        mock_mongo_hook.return_value.find.return_value = iter(MOCK_MONGO_RETURN)
        mock_s3_hook.return_value.load_string.return_value = True

        operator.execute(None)

        mock_mongo_hook.return_value.find.assert_called_once_with(
            mongo_collection=MONGO_COLLECTION,
            query=MONGO_QUERY,
            find_one=False,
            mongo_db=None,
            projection=None,
        )

        op_stringify = self.mock_operator._stringify
        op_transform = self.mock_operator.transform

        s3_doc_str = op_stringify(op_transform(MOCK_MONGO_RETURN))

        mock_s3_hook.return_value.load_string.assert_called_once_with(
            string_data=s3_doc_str, key=S3_KEY, bucket_name=S3_BUCKET, replace=False, compression=COMPRESSION
        )

    @mock.patch("airflow.providers.amazon.aws.transfers.mongo_to_s3.MongoHook")
    @mock.patch("airflow.providers.amazon.aws.transfers.mongo_to_s3.S3Hook")
    def test_execute_compress(self, mock_s3_hook, mock_mongo_hook):
        operator = self.mock_operator
        self.mock_operator.compression = "gzip"
        mock_mongo_hook.return_value.find.return_value = iter(MOCK_MONGO_RETURN)
        mock_s3_hook.return_value.load_string.return_value = True

        operator.execute(None)

        mock_mongo_hook.return_value.find.assert_called_once_with(
            mongo_collection=MONGO_COLLECTION,
            query=MONGO_QUERY,
            find_one=False,
            mongo_db=None,
            projection=None,
        )

        op_stringify = self.mock_operator._stringify
        op_transform = self.mock_operator.transform

        s3_doc_str = op_stringify(op_transform(MOCK_MONGO_RETURN))

        mock_s3_hook.return_value.load_string.assert_called_once_with(
            string_data=s3_doc_str, key=S3_KEY, bucket_name=S3_BUCKET, replace=False, compression="gzip"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.operators.eventbridge import (
    EventBridgeDisableRuleOperator,
    EventBridgeEnableRuleOperator,
    EventBridgePutEventsOperator,
    EventBridgePutRuleOperator,
)
from airflow.providers.common.compat.sdk import DAG, chain

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_eventbridge"
ENTRIES = [
    {
        "Detail": '{"event-name": "custom-event"}',
        "EventBusName": "custom-bus",
        "Source": "example.myapp",
        "DetailType": "Sample Custom Event",
    }
]


sys_test_context_task = SystemTestContextBuilder().build()

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    env_id = test_context[ENV_ID_KEY]

    # [START howto_operator_eventbridge_put_events]
    put_events = EventBridgePutEventsOperator(task_id="put_events_task", entries=ENTRIES)
    # [END howto_operator_eventbridge_put_events]

    # [START howto_operator_eventbridge_put_rule]
    put_rule = EventBridgePutRuleOperator(
        task_id="put_rule_task",
        name="example_rule",
        event_pattern='{"source": ["example.myapp"]}',
        description="This rule matches events from example.myapp.",
        state="DISABLED",
    )
    # [END howto_operator_eventbridge_put_rule]

    # [START howto_operator_eventbridge_enable_rule]
    enable_rule = EventBridgeEnableRuleOperator(task_id="enable_rule_task", name="example_rule")
    # [END howto_operator_eventbridge_enable_rule]

    # [START howto_operator_eventbridge_disable_rule]
    disable_rule = EventBridgeDisableRuleOperator(
        task_id="disable_rule_task",
        name="example_rule",
    )
    # [END howto_operator_eventbridge_disable_rule]

    chain(test_context, put_events, put_rule, enable_rule, disable_rule)


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from unittest import mock

import pytest

from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.amazon.aws.executors.batch.utils import (
    CONFIG_DEFAULTS,
    CONFIG_GROUP_NAME,
    AllBatchConfigKeys,
    BatchExecutorException,
    BatchJob,
    BatchJobCollection,
    BatchJobInfo,
    BatchQueuedJob,
    BatchSubmitJobKwargsConfigKeys,
)
from airflow.utils.state import State


class TestBatchQueuedJob:
    """Tests for the BatchQueuedJob dataclass."""

    def test_batch_queued_job_creation(self):
        """Test BatchQueuedJob object creation."""
        key = mock.Mock(spec=TaskInstanceKey)
        command = ["airflow", "tasks", "run"]
        queue = "default_queue"
        executor_config = {"key": "value"}
        attempt_number = 1
        next_attempt_time = datetime.datetime.now()

        queued_job = BatchQueuedJob(
            key=key,
            command=command,
            queue=queue,
            executor_config=executor_config,
            attempt_number=attempt_number,
            next_attempt_time=next_attempt_time,
        )

        assert queued_job.key is key
        assert queued_job.command == command
        assert queued_job.queue == queue
        assert queued_job.executor_config == executor_config
        assert queued_job.attempt_number == attempt_number
        assert queued_job.next_attempt_time is next_attempt_time


class TestBatchJobInfo:
    """Tests for the BatchJobInfo dataclass."""

    def test_batch_job_info_creation(self):
        """Test BatchJobInfo object creation."""
        cmd = ["airflow", "tasks", "run"]
        queue = "default_queue"
        config = {"key": "value"}

        job_info = BatchJobInfo(cmd=cmd, queue=queue, config=config)

        assert job_info.cmd == cmd
        assert job_info.queue == queue
        assert job_info.config == config


class TestBatchJob:
    """Tests for the BatchJob class."""

    @pytest.mark.parametrize(
        ("batch_status", "expected_airflow_state"),
        [
            ("SUBMITTED", State.QUEUED),
            ("PENDING", State.QUEUED),
            ("RUNNABLE", State.QUEUED),
            ("STARTING", State.QUEUED),
            ("RUNNING", State.RUNNING),
            ("SUCCEEDED", State.SUCCESS),
            ("FAILED", State.FAILED),
            ("UNKNOWN_STATUS", State.QUEUED),  # Default case
        ],
    )
    def test_get_job_state_mappings(self, batch_status, expected_airflow_state):
        """Test job state mappings from AWS Batch status to Airflow state."""
        job = BatchJob(job_id="job_id_123", status=batch_status)
        assert job.get_job_state() == expected_airflow_state

    def test_repr_method(self):
        """Test the __repr__ method for a meaningful string representation."""
        job_id = "test-job-123"
        status = "RUNNING"
        job = BatchJob(job_id=job_id, status=status)
        expected_repr = f"({job_id} -> {status}, {State.RUNNING})"
        assert repr(job) == expected_repr

    def test_status_reason_initialization(self):
        """Test BatchJob initialization with an optional status_reason."""
        job_id = "job-456"
        status = "FAILED"
        status_reason = "Insufficient resources"
        job = BatchJob(job_id=job_id, status=status, status_reason=status_reason)
        assert job.job_id == job_id
        assert job.status == status
        assert job.status_reason == status_reason

    def test_status_reason_default(self):
        """Test BatchJob initialization with a default status_reason."""
        job_id = "job-789"
        status = "SUCCEEDED"
        job = BatchJob(job_id=job_id, status=status)
        assert job.job_id == job_id
        assert job.status == status
        assert job.status_reason is None


class TestBatchJobCollection:
    """Tests for the BatchJobCollection class."""

    @pytest.fixture(autouse=True)
    def _setup_collection(self):
        """Set up a BatchJobCollection for testing."""
        self.collection = BatchJobCollection()
        self.key1 = mock.Mock(spec=TaskInstanceKey)
        self.key2 = mock.Mock(spec=TaskInstanceKey)
        self.job_id1 = "batch-job-001"
        self.job_id2 = "batch-job-002"
        self.cmd1 = ["command1"]
        self.cmd2 = ["command2"]
        self.queue1 = "queue1"
        self.queue2 = "queue2"
        self.config1 = {"conf1": "val1"}
        self.config2 = {"conf2": "val2"}

    def test_add_job(self):
        """Test adding a job to the collection."""
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=1,
        )
        assert len(self.collection) == 1
        assert self.collection.key_to_id[self.key1] == self.job_id1
        assert self.collection.id_to_key[self.job_id1] == self.key1
        assert self.collection.failure_count_by_id(self.job_id1) == 1
        assert self.collection.id_to_job_info[self.job_id1].cmd == self.cmd1
        assert self.collection.id_to_job_info[self.job_id1].queue == self.queue1
        assert self.collection.id_to_job_info[self.job_id1].config == self.config1

    def test_add_multiple_jobs(self):
        """Test adding multiple jobs to the collection."""
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=1,
        )
        self.collection.add_job(
            job_id=self.job_id2,
            airflow_task_key=self.key2,
            airflow_cmd=self.cmd2,
            queue=self.queue2,
            exec_config=self.config2,
            attempt_number=2,
        )
        assert len(self.collection) == 2
        assert self.collection.key_to_id[self.key1] == self.job_id1
        assert self.collection.key_to_id[self.key2] == self.job_id2
        assert self.collection.failure_count_by_id(self.job_id1) == 1
        assert self.collection.failure_count_by_id(self.job_id2) == 2

    def test_pop_by_id(self):
        """Test removing a job from the collection by its ID."""
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=1,
        )
        assert len(self.collection) == 1
        popped_key = self.collection.pop_by_id(self.job_id1)
        assert popped_key is self.key1
        assert len(self.collection) == 0
        assert self.job_id1 not in self.collection.id_to_key
        assert self.key1 not in self.collection.key_to_id
        # id_to_job_info is NOT removed by pop_by_id in the current implementation.
        assert self.job_id1 in self.collection.id_to_job_info
        assert self.collection.id_to_job_info[self.job_id1].cmd == self.cmd1
        # id_to_failure_counts is a defaultdict, so accessing a removed key returns 0.
        assert self.collection.id_to_failure_counts[self.job_id1] == 0

    def test_pop_non_existent_job_id(self):
        """Test popping a job ID that does not exist."""
        with pytest.raises(KeyError):
            self.collection.pop_by_id("non-existent-job-id")

    def test_failure_count_by_id(self):
        """Test getting failure count for a job ID."""
        attempt_number = 5
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=attempt_number,
        )
        assert self.collection.failure_count_by_id(self.job_id1) == attempt_number

    def test_failure_count_non_existent_job_id(self):
        """Test getting failure count for a non-existent job ID."""
        # id_to_failure_counts is a defaultdict, so it returns 0 for missing keys.
        assert self.collection.failure_count_by_id("non-existent-job-id") == 0

    def test_increment_failure_count(self):
        """Test incrementing the failure count for a job ID."""
        initial_attempt = 1
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=initial_attempt,
        )
        assert self.collection.failure_count_by_id(self.job_id1) == initial_attempt
        self.collection.increment_failure_count(self.job_id1)
        assert self.collection.failure_count_by_id(self.job_id1) == initial_attempt + 1

    def test_increment_failure_count_non_existent_job_id(self):
        """Test incrementing failure count for a non-existent job ID."""
        # id_to_failure_counts is a defaultdict, so incrementing a missing key sets it to 1.
        self.collection.increment_failure_count("non-existent-job-id")
        assert self.collection.failure_count_by_id("non-existent-job-id") == 1

    def test_get_all_jobs_empty(self):
        """Test getting all job IDs from an empty collection."""
        assert self.collection.get_all_jobs() == []

    def test_get_all_jobs_with_jobs(self):
        """Test getting all job IDs from a collection with jobs."""
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=1,
        )
        self.collection.add_job(
            job_id=self.job_id2,
            airflow_task_key=self.key2,
            airflow_cmd=self.cmd2,
            queue=self.queue2,
            exec_config=self.config2,
            attempt_number=1,
        )
        all_jobs = self.collection.get_all_jobs()
        assert len(all_jobs) == 2
        assert self.job_id1 in all_jobs
        assert self.job_id2 in all_jobs

    def test_len_method(self):
        """Test the __len__ method of the collection."""
        assert len(self.collection) == 0
        self.collection.add_job(
            job_id=self.job_id1,
            airflow_task_key=self.key1,
            airflow_cmd=self.cmd1,
            queue=self.queue1,
            exec_config=self.config1,
            attempt_number=1,
        )
        assert len(self.collection) == 1
        self.collection.add_job(
            job_id=self.job_id2,
            airflow_task_key=self.key2,
            airflow_cmd=self.cmd2,
            queue=self.queue2,
            exec_config=self.config2,
            attempt_number=1,
        )
        assert len(self.collection) == 2
        self.collection.pop_by_id(self.job_id1)
        assert len(self.collection) == 1


class TestConfigKeys:
    """Tests for configuration key constants."""

    def test_batch_submit_job_kwargs_config_keys_values(self):
        """Test BatchSubmitJobKwargsConfigKeys have correct string values."""
        assert BatchSubmitJobKwargsConfigKeys.JOB_NAME == "job_name"
        assert BatchSubmitJobKwargsConfigKeys.JOB_QUEUE == "job_queue"
        assert BatchSubmitJobKwargsConfigKeys.JOB_DEFINITION == "job_definition"
        assert BatchSubmitJobKwargsConfigKeys.EKS_PROPERTIES_OVERRIDE == "eks_properties_override"
        assert BatchSubmitJobKwargsConfigKeys.NODE_OVERRIDE == "node_override"

    def test_all_batch_config_keys_values(self):
        """Test AllBatchConfigKeys have correct string values, including inherited ones."""
        # Test keys specific to AllBatchConfigKeys
        assert AllBatchConfigKeys.MAX_SUBMIT_JOB_ATTEMPTS == "max_submit_job_attempts"
        assert AllBatchConfigKeys.AWS_CONN_ID == "conn_id"
        assert AllBatchConfigKeys.SUBMIT_JOB_KWARGS == "submit_job_kwargs"
        assert AllBatchConfigKeys.REGION_NAME == "region_name"
        assert AllBatchConfigKeys.CHECK_HEALTH_ON_STARTUP == "check_health_on_startup"

        # Test inherited keys from BatchSubmitJobKwargsConfigKeys
        assert AllBatchConfigKeys.JOB_NAME == "job_name"
        assert AllBatchConfigKeys.JOB_QUEUE == "job_queue"
        assert AllBatchConfigKeys.JOB_DEFINITION == "job_definition"
        assert AllBatchConfigKeys.EKS_PROPERTIES_OVERRIDE == "eks_properties_override"
        assert AllBatchConfigKeys.NODE_OVERRIDE == "node_override"

    def test_config_defaults(self):
        """Test that CONFIG_DEFAULTS is a dictionary with expected keys and values."""
        assert isinstance(CONFIG_DEFAULTS, dict)
        assert "conn_id" in CONFIG_DEFAULTS
        assert CONFIG_DEFAULTS["conn_id"] == "aws_default"
        assert "max_submit_job_attempts" in CONFIG_DEFAULTS
        assert CONFIG_DEFAULTS["max_submit_job_attempts"] == "3"
        assert "check_health_on_startup" in CONFIG_DEFAULTS
        assert CONFIG_DEFAULTS["check_health_on_startup"] == "True"

    def test_config_group_name(self):
        """Test that CONFIG_GROUP_NAME is a string."""
        assert isinstance(CONFIG_GROUP_NAME, str)
        assert CONFIG_GROUP_NAME == "aws_batch_executor"


class TestBatchExecutorException:
    """Tests for the BatchExecutorException class."""

    def test_exception_inheritance(self):
        """Test that BatchExecutorException is a subclass of Exception."""
        assert issubclass(BatchExecutorException, Exception)

    def test_exception_can_be_raised_and_caught(self):
        """Test that the exception can be raised and caught."""
        with pytest.raises(BatchExecutorException):
            raise BatchExecutorException("Test exception message")

    def test_exception_message(self):
        """Test that the exception stores the message correctly."""
        error_message = "An unexpected error occurred in the AWS Batch ecosystem."
        exc = BatchExecutorException(error_message)
        assert str(exc) == error_message



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.athena import AthenaHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.sdk import Context


class AthenaSensor(AwsBaseSensor[AthenaHook]):
    """
    Poll the state of the Query until it reaches a terminal state; fails if the query fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:AthenaSensor`


    :param query_execution_id: query_execution_id to check the state of
    :param max_retries: Number of times to poll for query state before
        returning the current state, defaults to None
    :param sleep_time: Time in seconds to wait between two consecutive call to
        check query status on athena, defaults to 10
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES = (
        "QUEUED",
        "RUNNING",
    )
    FAILURE_STATES = (
        "FAILED",
        "CANCELLED",
    )
    SUCCESS_STATES = ("SUCCEEDED",)

    aws_hook_class = AthenaHook
    template_fields: Sequence[str] = aws_template_fields(
        "query_execution_id",
    )
    ui_color = "#66c3ff"

    def __init__(
        self,
        *,
        query_execution_id: str,
        max_retries: int | None = None,
        sleep_time: int = 10,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.query_execution_id = query_execution_id
        self.sleep_time = sleep_time
        self.max_retries = max_retries

    def poke(self, context: Context) -> bool:
        state = self.hook.poll_query_status(self.query_execution_id, self.max_retries, self.sleep_time)

        if state in self.FAILURE_STATES:
            raise AirflowException("Athena sensor failed")

        if state in self.INTERMEDIATE_STATES:
            return False
        return True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Module for suppress errors in Amazon Provider.

.. warning::
    Only for internal usage, this module might be changed or removed in the future
    without any further notice.

:meta: private
"""

from __future__ import annotations

import logging
from collections.abc import Callable
from functools import wraps
from typing import ParamSpec, TypeVar

PS = ParamSpec("PS")
RT = TypeVar("RT")

log = logging.getLogger(__name__)


def return_on_error(return_value: RT):
    """
    Suppress any ``Exception`` raised in decorator function.

    Main use-case when functional is optional, however any error on functions/methods might
    raise any error which are subclass of ``Exception``.

    .. note::
        Decorator doesn't intend to catch ``BaseException``,
        e.g. ``GeneratorExit``, ``KeyboardInterrupt``, ``SystemExit`` and others.

    .. warning::
        Only for internal usage, this decorator might be changed or removed in the future
        without any further notice.

    :param return_value: Return value if decorated function/method raise any ``Exception``.
    :meta: private
    """

    def decorator(func: Callable[PS, RT]) -> Callable[PS, RT]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> RT:
            try:
                return func(*args, **kwargs)
            except Exception:
                log.debug(
                    "Encountered error during execution function/method %r", func.__name__, exc_info=True
                )
                return return_value

        return wrapper

    return decorator



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.glue_crawler import GlueCrawlerHook
from airflow.providers.amazon.aws.triggers.glue_crawler import GlueCrawlerCompleteTrigger
from airflow.triggers.base import TriggerEvent

from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type


class TestGlueCrawlerCompleteTrigger:
    def test_serialization(self):
        crawler_name = "test_crawler"
        poll_interval = 10
        aws_conn_id = "aws_default"

        trigger = GlueCrawlerCompleteTrigger(
            crawler_name=crawler_name,
            waiter_delay=poll_interval,
            aws_conn_id=aws_conn_id,
        )

        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.glue_crawler.GlueCrawlerCompleteTrigger"
        assert kwargs == {
            "crawler_name": "test_crawler",
            "waiter_delay": 10,
            "waiter_max_attempts": 1500,
            "aws_conn_id": "aws_default",
        }

    @pytest.mark.asyncio
    @mock.patch.object(GlueCrawlerHook, "get_waiter")
    @mock.patch.object(GlueCrawlerHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        crawler_name = "test_crawler"
        trigger = GlueCrawlerCompleteTrigger(crawler_name=crawler_name)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "value": None})
        assert_expected_waiter_type(mock_get_waiter, "crawler_ready")
        mock_get_waiter().wait.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator
from functools import cached_property
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook

from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook, GlueJobHook
from airflow.providers.amazon.aws.hooks.glue_catalog import GlueCatalogHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger
from airflow.triggers.base import BaseTrigger, TriggerEvent


class GlueJobCompleteTrigger(AwsBaseWaiterTrigger):
    """
    Watches for a glue job, triggers when it finishes.

    :param job_name: glue job name
    :param run_id: the ID of the specific run to watch for that job
    :param verbose: whether to print the job's logs in airflow logs or not
    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 60)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials
    :param region_name: Optional aws region name (example: us-east-1). Uses region from connection
        if not specified.
    :param verify: Whether or not to verify SSL certificates.
    :param botocore_config: Configuration dictionary (key-values) for botocore client.
    """

    def __init__(
        self,
        job_name: str,
        run_id: str,
        verbose: bool = False,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = "aws_default",
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
    ):
        super().__init__(
            serialized_fields={"job_name": job_name, "run_id": run_id, "verbose": verbose},
            waiter_name="job_complete",
            waiter_args={"JobName": job_name, "RunId": run_id},
            failure_message="AWS Glue job failed.",
            status_message="Status of AWS Glue job is",
            status_queries=["JobRun.JobRunState"],
            return_key="run_id",
            return_value=run_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
            verify=verify,
            botocore_config=botocore_config,
        )
        self.job_name = job_name
        self.run_id = run_id
        self.verbose = verbose

    def hook(self) -> AwsGenericHook:
        return GlueJobHook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )


class GlueCatalogPartitionTrigger(BaseTrigger):
    """
    Asynchronously waits for a partition to show up in AWS Glue Catalog.

    :param database_name: The name of the catalog database where the partitions reside.
    :param table_name: The name of the table to wait for, supports the dot
        notation (my_database.my_table)
    :param expression: The partition clause to wait for. This is passed as
        is to the AWS Glue Catalog API's get_partitions function,
        and supports SQL like notation as in ``ds='2015-01-01'
        AND type='value'`` and comparison operators as in ``"ds>=2015-01-01"``.
        See https://docs.aws.amazon.com/glue/latest/dg/aws-glue-api-catalog-partitions.html
        #aws-glue-api-catalog-partitions-GetPartitions
    :param aws_conn_id: ID of the Airflow connection where
        credentials and extra configuration are stored
    :param region_name: Optional aws region name (example: us-east-1). Uses region from connection
        if not specified.
    :param waiter_delay: Number of seconds to wait between two checks. Default is 60 seconds.
    """

    def __init__(
        self,
        database_name: str,
        table_name: str,
        expression: str = "",
        waiter_delay: int = 60,
        aws_conn_id: str | None = "aws_default",
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
    ):
        self.database_name = database_name
        self.table_name = table_name
        self.expression = expression
        self.waiter_delay = waiter_delay

        self.aws_conn_id = aws_conn_id
        self.region_name = region_name
        self.verify = verify
        self.botocore_config = botocore_config

    def serialize(self) -> tuple[str, dict[str, Any]]:
        return (
            # dynamically generate the fully qualified name of the class
            self.__class__.__module__ + "." + self.__class__.__qualname__,
            {
                "database_name": self.database_name,
                "table_name": self.table_name,
                "expression": self.expression,
                "aws_conn_id": self.aws_conn_id,
                "region_name": self.region_name,
                "waiter_delay": self.waiter_delay,
                "verify": self.verify,
                "botocore_config": self.botocore_config,
            },
        )

    @cached_property
    def hook(self) -> GlueCatalogHook:
        return GlueCatalogHook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )

    async def poke(self, client: Any) -> bool:
        if "." in self.table_name:
            self.database_name, self.table_name = self.table_name.split(".")
        self.log.info(
            "Poking for table %s. %s, expression %s", self.database_name, self.table_name, self.expression
        )
        partitions = await self.hook.async_get_partitions(
            client=client,
            database_name=self.database_name,
            table_name=self.table_name,
            expression=self.expression,
        )

        return bool(partitions)

    async def run(self) -> AsyncIterator[TriggerEvent]:
        async with await self.hook.get_async_conn() as client:
            while True:
                result = await self.poke(client=client)
                if result:
                    yield TriggerEvent({"status": "success"})
                    break
                else:
                    await asyncio.sleep(self.waiter_delay)


class GlueDataQualityRuleSetEvaluationRunCompleteTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a AWS Glue data quality evaluation run complete.

    :param evaluation_run_id: The AWS Glue data quality ruleset evaluation run identifier.
    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 60)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        evaluation_run_id: str,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = "aws_default",
    ):
        super().__init__(
            serialized_fields={"evaluation_run_id": evaluation_run_id},
            waiter_name="data_quality_ruleset_evaluation_run_complete",
            waiter_args={"RunId": evaluation_run_id},
            failure_message="AWS Glue data quality ruleset evaluation run failed.",
            status_message="Status of AWS Glue data quality ruleset evaluation run is",
            status_queries=["Status"],
            return_key="evaluation_run_id",
            return_value=evaluation_run_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return GlueDataQualityHook(aws_conn_id=self.aws_conn_id)


class GlueDataQualityRuleRecommendationRunCompleteTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a AWS Glue data quality recommendation run complete.

    :param recommendation_run_id: The AWS Glue data quality rule recommendation run identifier.
    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 60)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        recommendation_run_id: str,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = "aws_default",
    ):
        super().__init__(
            serialized_fields={"recommendation_run_id": recommendation_run_id},
            waiter_name="data_quality_rule_recommendation_run_complete",
            waiter_args={"RunId": recommendation_run_id},
            failure_message="AWS Glue data quality recommendation run failed.",
            status_message="Status of AWS Glue data quality recommendation run is",
            status_queries=["Status"],
            return_key="recommendation_run_id",
            return_value=recommendation_run_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return GlueDataQualityHook(aws_conn_id=self.aws_conn_id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.opensearch_serverless import OpenSearchServerlessHook
from airflow.providers.amazon.aws.triggers.opensearch_serverless import (
    OpenSearchServerlessCollectionActiveTrigger,
)
from airflow.triggers.base import TriggerEvent
from airflow.utils.helpers import prune_dict

from unit.amazon.aws.triggers.test_base import TestAwsBaseWaiterTrigger

BASE_TRIGGER_CLASSPATH = "airflow.providers.amazon.aws.triggers.opensearch_serverless."


class TestBaseBedrockTrigger(TestAwsBaseWaiterTrigger):
    EXPECTED_WAITER_NAME: str | None = None

    def test_setup(self):
        # Ensure that all subclasses have an expected waiter name set.
        if self.__class__.__name__ != "TestBaseBedrockTrigger":
            assert isinstance(self.EXPECTED_WAITER_NAME, str)


class TestOpenSearchServerlessCollectionActiveTrigger:
    EXPECTED_WAITER_NAME = "collection_available"
    COLLECTION_NAME = "test_collection_name"
    COLLECTION_ID = "test_collection_id"

    @pytest.mark.parametrize(
        ("collection_name", "collection_id", "expected_pass"),
        [
            pytest.param(COLLECTION_NAME, COLLECTION_ID, False, id="both_provided_fails"),
            pytest.param(COLLECTION_NAME, None, True, id="only_name_provided_passes"),
            pytest.param(None, COLLECTION_ID, True, id="only_id_provided_passes"),
        ],
    )
    def test_serialization(self, collection_name, collection_id, expected_pass):
        """Assert that arguments and classpath are correctly serialized."""
        call_args = prune_dict({"collection_id": collection_id, "collection_name": collection_name})

        if expected_pass:
            trigger = OpenSearchServerlessCollectionActiveTrigger(**call_args)
            classpath, kwargs = trigger.serialize()
            assert classpath == BASE_TRIGGER_CLASSPATH + "OpenSearchServerlessCollectionActiveTrigger"
            if call_args.get("collection_name"):
                assert kwargs.get("collection_name") == self.COLLECTION_NAME
            if call_args.get("collection_id"):
                assert kwargs.get("collection_id") == self.COLLECTION_ID

        if not expected_pass:
            with pytest.raises(
                AttributeError, match="Either collection_ids or collection_names must be provided, not both."
            ):
                OpenSearchServerlessCollectionActiveTrigger(**call_args)

    @pytest.mark.asyncio
    @mock.patch.object(OpenSearchServerlessHook, "get_waiter")
    @mock.patch.object(OpenSearchServerlessHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = OpenSearchServerlessCollectionActiveTrigger(collection_id=self.COLLECTION_ID)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "collection_id": self.COLLECTION_ID})
        assert mock_get_waiter().wait.call_count == 1
